/* programme démonstrateur pour un dialogue série par l'intermédiaire des deux lignes filaires
   TX et RX des cartes Arduino NANO. Le programme Maître sera téléchagé ultérieurement sur la carte NANO
   de la Raquette de télécommande.
   
 La procedure Attendre_une_chaine_sur_le_Maitre() est différente de celle de l'Esclave car elle doit
 sauter le premier délimiteur '!' pour détecter la sentinelle qui est le deuxième '!'. Ce protocole
 suppose que l'Esclave qui équipe la sonde ne retourne que des ACR ENCADRÉS par des '!'.
 
  UTILISATION DU DEMONSTRATEUR :
  ------------------------------
 Ce programme est intimement lié au programme P19_Esclave_minimal. Les deux circuits sont reliés
 en mode filaire croisé sur les lignes TX et RX. Ce programme permet d'envoyer comme consigne "a*"
 si on clique sur le BP "du "haut"" et "l*" sur le B.P. du "bas. Puis il attend l'ACR de l'esclave.
 Si l'ACR est "!OK!" il allume la LED sur D4. Si l'ACR est "!ERR!" il l'éteint. Même comportement
 sur la LED en D7 de l'Esclave. Plusieurs clics sur un B.P. semblent sans effet puisqu'il n'y a pas
 changement d'état des LEDs Raison pour laquelle le BIP sur le BUZZER permet de vérifier qu'il y a
 eu un alternat correct. Les lesttre "A" et "L" pour les deux consignes permettent de brancher ce
 programme démonstrateur Maître sur le programme P30 non modifié. Les B.P. permettent d'allumer et
 déteindre les phares et le LASER. Il faut dans ce but relier TX et RX en mode croiré sur D0 et D1.
 
 DIFFERENTS BRANCHEMENTS :
 A0 : Clavier provisoire à deux boutons câblés en analogique.
 D0 :  RX > Ligne série de dialogue va à TX du Maître.
 D1 :  TX > Ligne série de dialogue va à RX du Maître.
 D4 : LED témoin du dialogue entre machines.
 
                                                             PGM               RAM dynamique 
 Programme démonstrateur Maître minimal :               4518 Octets (14%)    222 Octets (10%)
*/
   
//--------------------- Déclarations pour le matériel -----------------------

//-------------- Définition des broches de sorties Binaires -----------------
#define LED_Dialogue 4

//------------------------ Constantes du programme --------------------------
#define LGR_chaine 15 // Longueur maximale de la chaine de dialogue.
#define NB_tests_antirebonds 50 // Nombre de tests de stabilisation B.P.

//------------------------- Variables du programme --------------------------
// Note : Les déclarations sont classées par type et par ordre alphabétique croissant.
boolean Augmenter_le_choix; // gestion des deux B.P.
boolean Clic_sur_le_BP; // gestion des deux B.P.
boolean Premier_delimiteur_OK; // La sentinelle sera le deuxième délimiteur '!'.
boolean Sentinelle_detectee;  // Délimiteur de fin détecté.
char Caractere;
byte Compteur_caracteres;
String Chaine_Memorisee; // Tampon mémoire pour la chaine entrante.

void setup()  {
  Serial.begin(115200);
  Chaine_Memorisee.reserve(LGR_chaine); // Reserve LGR_chaine octets pour la chaine.
  pinMode(LED_Dialogue, OUTPUT);}

void loop() {
  Tester_entree_A0(); // Pour détection dans la boucle de base.
  // ========= Traiter la requête de l'un des deux B.P. ========
  if (Clic_sur_le_BP) {
    Clic_sur_le_BP = false; // Action sur le B.P. prise en compte.
    if (Augmenter_le_choix) Serial.print("a*");
    else Serial.print("l*");}
  //-------------- Traite un ACR arrivée sur USB ---------------
  if (Serial.available()) {
      Attendre_une_chaine_sur_le_Maitre();
      if (Chaine_Memorisee == "!OK!") digitalWrite(LED_Dialogue,HIGH);
      if (Chaine_Memorisee == "!ERR!") digitalWrite(LED_Dialogue,LOW);} }
  
void Attendre_une_chaine_sur_le_Maitre() {// Cette procédure attend une chaine sur la ligne USB.
//-------------- Définition des broches de sorties Binaires -----------------
  Chaine_Memorisee = ""; // Vider le tampon pour une nouvelle capture.
  Sentinelle_detectee = false;  // Sentinelle non reçue.
  Premier_delimiteur_OK = false; // Premier '!' non encore reçu.
  Compteur_caracteres=0;
  while (!Sentinelle_detectee) {
        if (Serial.available()) {Caractere = Serial.read(); // Puiser un caractère dans le FIFO.
        // Puis concaténer ce caractère y compris la sentinelle :
        if (Compteur_caracteres < LGR_chaine) 
            {Chaine_Memorisee = Chaine_Memorisee + Caractere; Compteur_caracteres++;}
        // Si sentinelle détectée, informer la boucle de réception :
        Sentinelle_detectee = ((Caractere == '!') && (Premier_delimiteur_OK));
        Premier_delimiteur_OK = true; }}} // On saute le premier '!'. 

// >>>>>>>>>>>>> Le B.P. est câblé en "Analogique" <<<<<<<<<<<<<<<<<<
void Tester_entree_A0() { // La lecture sur A0 donne 0 ou 930 maximum.
  int Numerisation_BP; // Valeur retournée par la CAN de A0.
  byte COMPTEUR; unsigned long Horloge_actuelle; 
  Numerisation_BP = analogRead(0);
  if ((Numerisation_BP > 710) || (Numerisation_BP < 240)) {
     Clic_sur_le_BP = true;
     Augmenter_le_choix = false;
     if (Numerisation_BP > 710) {Augmenter_le_choix = true;}
     delay(100); // Attendre_la_stabilisation du travail.
     // Attendre_la_stabilisation du repos :
     COMPTEUR = 0;
     while (COMPTEUR < NB_tests_antirebonds) {
        COMPTEUR++; Numerisation_BP = analogRead(0);
        if ((Numerisation_BP > 710) || (Numerisation_BP < 240)) COMPTEUR = 0; } } }
