/* Programme Démonstrateur pour animer la Sonde JEKERT.
   Programme à associer avec P21B_Demonstrateur_Raquette.ino
   
   UTILISATION DU PROGRAMME :
 ----------------------------
 1) Établir la liaison TX/RX avec l'inverseur,
 2) Faire un RESET sur la Raquette,
 3) Valider l'éditeur de P22B_Demonstrateur_Sonde.ino sur le P.C.
 4) Activer le Moniteur de l'IDE. (Le P.C. affichera les ACR de la sonde, donc les n°ERR et la version.
 5) Cliquer à convenance sur toutes les touches du clavier.
    Si ERR ou version du PGM un BIP est déclenché.
 6) Cliquer sur le bouton central du codeur incrémental se contente de générer un BIP.
 7) Tourner à convenance le codeur incrémental. Il n'y aura un effet sur le répétiteur, sur le LASER et
    sur les Phares que si une CIBLE est préalablement désignée avec S13 ou S14. Les débordements
    dans un sens ou dans l'autre génèrent des erreurs n°1 ou n°2.
 Pour tester la LED rouge BUZZER sur la sonde, S9 et S10 de Rapide et Lent alternent son état.
   
 Ce que fait ce démonstrateur :
 • Il traite les consignes 24 à 36 de P21A_Demonstrateur_Raquette.ino.
 • Il traite les consignes 40 et 43 pour le gradateur lumineux LASER et Phares.
 • Gère la LED jaune pour débordement POSITIF du codeur rotatif.
 • Gère PROVISOIREMENR la LED jaune pour débordement NEGATIF du codeur rotatif.
 • Mise en place du booléen APPRENTISSAGE.
 • Intégration de l'ACR pour retourner l'état des "booléens machine".
 • Intégration de l'ACR pour retourner la valeur de la tension sur les servomoteurs.
 • Intégration de l'ACR pour retourner les valeurs des gradateurs Phares et LASER.

                                                             PGM               RAM dynamique 
Démonstrateur avant modifications :                    5598 Octets (18%)    282 Octets (13%)
Ajout du codeur rotatif, gradateur lumineux ... :      6112 Octets (19%)    259 Octets (12%)  +514
Booléen APPRENTISSAGE et ACR "état système" :          6460 Octets (21%)    264 Octets (12%)  +348
Retour valeur tension servomoteurs :                   8932 Octets (29%)    266 Octets (13%) +2472
Valeurs gradateurs Phares et LASER :                   9156 Octets (29%)    266 Octets (13%)  +224
Valeurs Hygrométrie et Température :                   9652 Octets (31%)    268 Octets (13%)  +496
Valeurs du luxmètre :                                  9836 Octets (32%)    268 Octets (13%)  +184
Télémètre à ultrasons :                               10238 Octets (33%)    269 Octets (13%)  +402
*/
   
//--------------------- Déclarations pour le matériel -----------------------
// Déclarations pour le capteur d'humidité DHT11.
#include <DHT11.h>
//--------------- Définition des broches d'Entrées Analogiques --------------
#define Entree_Luxmetre 6 // Entrée analogique A6 utilisée pour le luxmètre.
//--------------- Définition des broches d'Entrées BINAIRES -----------------
#define Bouclier_au_sol 3 // Entrée de détection du blouclier en contact avec le sol.
//------------------ Sorties sur broches "Analogiques" ----------------------
#define Entree_Alim_moteurs 7 // Entrée analogique A7 utilisée pour mesurer la tension de puissance.
#define LED_Rapide 14 // Entrée analogique A0 utilisée comme Sortie Binaire.
#define LED_Torsion_active 16 // Entrée analogique A2 utilisée comme Sortie Binaire.
#define LED_Loop_active 17 // Entrée analogique A3 utilisée comme Sortie Binaire.
//-------------- Définition des broches de sorties Binaires -----------------
#define BUZZER 2 // Le bruiteur est piloté par la sortie binaire D2.
#define Disjoncteur 4 // Masse des LEDs pour le mode Sommeil : Sortie binaire D4.
#define LED_Moteurs_OFF 7 // Pilotage LED témoin Moteurs figés.
#define Debordement_positif 8 // Pilotage LED de débordement positif du codeur rotatif.
#define Broche_Com_DH11 11 // Ligne de dialogue avec le capteur d'humidité DHT11.
#define TX_Pulse_sonard 12 // TX capteur ultrasons sur la sortie D12.
#define Echo 13 // RX capteur ultrasons sur la sortie D12.
//-------------- Broches de sorties Binaires utilisées en PWM ---------------
#define Pilotage_eclairage 5 // Pilotage des deux LEDs phares de la sonde : Sortie binaire D5.
#define LASER 6 // Sortie binaire D6 pour piloter le LASER en PWM.
#define Repetiteur_CAN 10 // Sortie D10 pour piloter la LED répétiteur di CAN.
//------------------------ Constantes du programme --------------------------
#define Version_PGM "> P22-B <" // @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
#define Calibre_U_moteurs 6.67 // Calibrage du diviseur potentiométrique d'entrée. @@@@@@@@@@@@@@@@@@@@@@@@@@@
#define LGR_chaine 4 // Longueur maximale de la chaine de dialogue : Quatre en réception.
#define LGR_ACR 13 // Longueur maximale de la chaine d'ACR sur TX. @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
// Codes d'erreurs de la fonction readDHT11() :
enum {
  DHT11_SUCCESS,          // 0 : Pas d'erreur.
  DHT11_TIMEOUT_ERROR,    // 1 : Temps d'attente dépassé.
  DHT11_CHECKSUM_ERROR};  // 2 : Données reçues erronées.

//------------------------- Variables du programme --------------------------
// Note : Les déclarations sont classées par type et par ordre alphabétique croissant.
boolean Apprentissage; // Mode apprentissage en cours.
boolean LASER_actif; // Allumer ou éteindre le LASER.
boolean Moteurs_ON; // Si false tous les moteurs sont figés.
boolean Phares_actifs; // Allumer ou éteindre les phares.
boolean Rapide; // Les mouvements coordonnés seront rapides.
boolean Repondre;
boolean Sentinelle_detectee;  // Délimiteur de fin détecté.
boolean Stabilisation_gyroscopique;
boolean Torsion_Active; // Les corrections en torsion sont actives.

char Caractere;
byte Cible_Codeur_incremental; // 1 : Phares, 2 : LASER.
byte Compteur_caracteres;
byte Distance; // Utilisée également pour la gestion de l'EEPROM.
byte Humidite; // Capteur d'humidité DHT11.
byte LSB; // Utilisé par le capteur magnétique.
byte Niveau_eclairage_LASER; // Pilotage analogique du LASER.
byte Niveau_eclairage_Phares; // Pilotage analogique des deux phares.
byte num_ERR;
byte Temperature; // Capteur d'humidité DHT11.
byte Valeur_Codeur_incremental;

int  NOMBRE; // Contiendra la transformée numérique d'une chaine.

unsigned long Ancienne_valeur_Temps_Ecoule_depuis_RESET;

String Chaine_Memorisee; // Tampon mémoire pour la chaine entrante.
String Texte_ACR; // Tampon mémoire pour la chaine sortante.

void setup()  {
  Serial.begin(115200); Repondre = false;
  Chaine_Memorisee.reserve(LGR_chaine); // Reserve LGR_chaine octets pour la chaine.
  Chaine_Memorisee.reserve(LGR_ACR); // Reserve LGR_chaine octets pour la chaine.
  //--------------------------- Initialiser les booléens ---------------------------
  Phares_actifs = false; LASER_actif = false; Stabilisation_gyroscopique = false; Rapide = true; 
  Torsion_Active = false; Moteurs_ON = true; // À enlever quand le module inertiel sera introduit @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
  Apprentissage = false;
  //--------------------------- Initialiser les sorties ----------------------------
  Niveau_eclairage_LASER = 127; Niveau_eclairage_Phares = 127;
  pinMode(Disjoncteur,OUTPUT); digitalWrite(Disjoncteur,HIGH); // Transistor saturé. (Masse active.)
  pinMode(LED_Loop_active,OUTPUT); pinMode(LED_Moteurs_OFF,OUTPUT); digitalWrite(LED_Moteurs_OFF,LOW);
  pinMode(LED_Torsion_active,OUTPUT); digitalWrite(LED_Torsion_active,LOW);
  pinMode(LED_Rapide,OUTPUT); digitalWrite(LED_Rapide,LOW); pinMode(Echo,INPUT);
  pinMode(TX_Pulse_sonard,OUTPUT); digitalWrite(TX_Pulse_sonard,LOW); // TX HC-SR04 ultrasons.
  pinMode(Debordement_positif,OUTPUT); digitalWrite(Debordement_positif,LOW);
  //--------------------------------------------------------------------------------
  Chaine_Memorisee = "";} // Vider le tampon pour une nouvelle capture.

void loop() {
  //--------------------- Traite une Consigne arrivée sur USB ----------------------
  if (Serial.available()) Attendre_une_chaine_sur_le_Maitre();
  //-------------------------- Traitement de la Consignee --------------------------
  if (Chaine_Memorisee.length() > 0) {Analyseur_Syntaxique(); Traite_la_consigne();}
  //-------------------- Traitement des actions de la boucle de base ---------------
  //-------- Faire clignoter la LED de la boucle de base active --------
  LSB = 100; if (Stabilisation_gyroscopique) LSB = 30;
  if (millis() > Ancienne_valeur_Temps_Ecoule_depuis_RESET) {
      Ancienne_valeur_Temps_Ecoule_depuis_RESET = millis() + LSB; // Réarme pour 0,1S ou 0,03S.
      digitalWrite(LED_Loop_active,!digitalRead(LED_Loop_active));} // Inverser l'état de la LED.
  //----------------- Traite les fonctions analogiques -----------------
  analogWrite(Repetiteur_CAN, Valeur_Codeur_incremental);
  if (Phares_actifs) analogWrite(Pilotage_eclairage, Niveau_eclairage_Phares);
  else analogWrite(Pilotage_eclairage, 0);
  if (LASER_actif) analogWrite(LASER, Niveau_eclairage_LASER);
  else analogWrite(LASER, 0);
  //--------------- Fin du traitement des actions de la boucle de base -------------
  if (Repondre) Accuse_reception();} // ACR si une consigne a été reçue.
      
void Accuse_reception() {      
  delay(100); Repondre = false; Serial.print(Texte_ACR); 
  if (num_ERR > 0) Serial.print(num_ERR); Serial.print('*');}
  
void Traite_la_consigne() {
  Texte_ACR = "OK";
  switch (NOMBRE) {
      case 24 : {Texte_ACR = Version_PGM; break;} // S16 Indique la version du programme.   
      case 25 : {digitalWrite(Disjoncteur,LOW); break;} // S1 Disjoncte les LEDs.
      case 26 : {digitalWrite(Disjoncteur,HIGH); break; } // S2 Rétablit les LEDs.
      case 27 : {Phares_actifs = true; break;} // S3 Allume les Phares.
      case 28 : {Phares_actifs = false; break;} // S4 Éteind les Phares.
      case 29 : {LASER_actif = true; break;} // S5 Allume le LASER.
      case 30 : {LASER_actif = false; break;} // S6 Éteind le LASER.
      case 31 : {Moteurs_ON = false; digitalWrite(LED_Moteurs_OFF,HIGH); break;} // S7 Figer les moteurs.
      case 32 : {Moteurs_ON = true; digitalWrite(LED_Moteurs_OFF,LOW); break;} // S8 Rétablir les moteurs.
      case 33 : {Rapide = true; digitalWrite(LED_Rapide,HIGH); break;} // S9 Mvt de coordination rapides.
      case 34 : {Rapide = false; digitalWrite(LED_Rapide,LOW); break;} // S10 Mvt de coordination lents.
      case 35 : {Stabilisation_gyroscopique = true; break;} // S11 Début de stabilisation gyroscopique.
      case 36 : {Stabilisation_gyroscopique = false; break;} // S12 Fin de stabilisation gyroscopique.
      case 37 : {Apprentissage = true; break;} // Pas de B.P. Inclus pour tester la version D.
      case 38 : {Apprentissage = false; break;} // Pas de B.P. Inclus pour tester la version D.
      case 39 : {Torsion_Active = false; digitalWrite(LED_Torsion_active,LOW); Texte_ACR = "> C39<" ; break;}  // S15 Terminer un mouvement de torsion. @@@@@@@
      case 40 : {Faire_un_pas_sens_Positif(); break;} // Codeur Incrémental Incrémenté.
      case 41 : {Faire_un_pas_sens_Negatif(); break;} // Codeur Incrémental Décrémenté.
      case 42 : {Change_cible_codeur_incremental(1); break;} // S13 Le codeur incrémental agit sur le LASER.
      case 43 : {Change_cible_codeur_incremental(2); break;} // S14 " " " " " " " " " agit sur les Phares.
      //--------- Codes 44 à 51 pas de touche clavier. Ajoutés pour tester la version D du Maitre.--------- 
      case 44 : {Construire_ACR_etat_machine(); break;}    
      case 45 : {Texte_ACR = String(analogRead(Entree_Alim_moteurs) * Calibre_U_moteurs / 1023); break;}
      case 46 : {Texte_ACR = String(Niveau_eclairage_Phares); break;}
      case 47 : {Texte_ACR = String(Niveau_eclairage_LASER); break;}
      case 48 : {Mesurer_Hygrometrie_et_Temperature();
                 if (num_ERR == 0) Texte_ACR = String(Temperature); else Texte_ACR = String(num_ERR); break;}
      case 49 : {Mesurer_Hygrometrie_et_Temperature();
                 if (num_ERR == 0) Texte_ACR = String(Humidite); else Texte_ACR = String(num_ERR); break;}
      case 50 : {Texte_ACR = String(analogRead(Entree_Luxmetre)); break;}
      case 51 : {Telemetre_ultrasons(); Texte_ACR = String(Distance); break;}}      
  if (num_ERR > 0) Texte_ACR = "ERR ";}
    
void Construire_ACR_etat_machine() {
 if (digitalRead(Disjoncteur)) Texte_ACR = "0"; else Texte_ACR = "1";
 Ajoute_booleen(Moteurs_ON); Ajoute_booleen(Rapide); Ajoute_booleen(Stabilisation_gyroscopique);
 if (digitalRead(Bouclier_au_sol)) Texte_ACR = Texte_ACR + "1"; else Texte_ACR = Texte_ACR + "0";
 Ajoute_booleen(Phares_actifs); Ajoute_booleen(LASER_actif); Ajoute_booleen(Apprentissage);}

void Ajoute_booleen(boolean ETAT) {
 if (ETAT) Texte_ACR = Texte_ACR + "1"; else Texte_ACR = Texte_ACR + "0";} 
     
void Faire_un_pas_sens_Positif() {
  switch (Cible_Codeur_incremental) {
    case 1 : {Augmente_energie(); Niveau_eclairage_LASER = Valeur_Codeur_incremental; break;}
    case 2 : {Augmente_energie(); Niveau_eclairage_Phares = Valeur_Codeur_incremental;}}
    if (num_ERR > 0) digitalWrite(Debordement_positif,HIGH); else digitalWrite(Debordement_positif,LOW);}
  
void Faire_un_pas_sens_Negatif() {
  switch (Cible_Codeur_incremental) {
    case 1 : {Diminue_energie(); Niveau_eclairage_LASER = Valeur_Codeur_incremental; break;}
    case 2 : {Diminue_energie(); Niveau_eclairage_Phares = Valeur_Codeur_incremental;}}
    if (num_ERR > 0) digitalWrite(Debordement_positif,HIGH);
    else digitalWrite(Debordement_positif,LOW);} // S13 en version définitive. @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

void Change_cible_codeur_incremental(byte CIBLE) {
  Cible_Codeur_incremental = CIBLE;
  switch (CIBLE) {
    case 1 : {Valeur_Codeur_incremental = Niveau_eclairage_LASER; break;} // Le codeur agit sur le LASER.
    case 2 : {Valeur_Codeur_incremental = Niveau_eclairage_Phares;}}} // // Le codeur agit sur les Phares.

void Augmente_energie() {
  if (Valeur_Codeur_incremental < 245) Valeur_Codeur_incremental = Valeur_Codeur_incremental + 9;
  else {num_ERR = 1; Valeur_Codeur_incremental = 254;}} // Ne pas dépasser 254.
  
void Diminue_energie() {
 if (Valeur_Codeur_incremental > 9) Valeur_Codeur_incremental = Valeur_Codeur_incremental - 9;
 else {num_ERR = 2; Valeur_Codeur_incremental = 1;}} // Ne pas descendre en dessous de 1.
 
//--------------------------- Télémétrie à capteur Ultrasons HC-SR04 ---------------------------
void Telemetre_ultrasons() {
  // Déclencher une mesure :
  digitalWrite(TX_Pulse_sonard, HIGH); delayMicroseconds(10); digitalWrite(TX_Pulse_sonard, LOW);
  // Analyser le retour sur la ligne de réception :
  Distance = pulseIn(Echo, HIGH) / 58;}
 
//----------------------------- Routines du capteur d'Humidité DHT11 ---------------------------
/* >>> Lecture de la température et lu taux d'humidité mesurés par un capteur DHT11 <<<
  * Broche_DATA : Broche sur laquelle est câblée la sortie du capteur.
  * Temperature : Pointeur vers la variable stockant la température.
  * Humidite : Pointeur vers la variable stockant le taux d'humidité.
  * Retourne DHT11_SUCCESS si aucune erreur, DHT11_TIMEOUT_ERROR en cas de dépassement,
    ou DHT11_CHECKSUM_ERROR en cas d'erreur de checksum */

void Mesurer_Hygrometrie_et_Temperature() {
  // >>> Lecture de la température et de l'humidité + gestion des erreurs <<<
  switch (readDHT11(Broche_Com_DH11, &Temperature, &Humidite)) {
    case DHT11_SUCCESS: break; //Ne rien faire si OK.          
    case DHT11_TIMEOUT_ERROR: {num_ERR = 3; break;}
    case DHT11_CHECKSUM_ERROR: num_ERR = 4;}} // Fin du case DHT11_SUCCESS:
   
byte readDHT11(byte Broche_DATA, byte* Temperature, byte* Humidite) {
    // Variable_DATA[] : Tampon contenant les données du cpateur.
    // BIT_serie : Compteur permettant de savoir quel bit est reçu. (bitwise)
    // Index : Compteur permettant de savoir quel octet est reçu. (bitwise)
    // timeout : Compteur pour le délai en dépassement.
  byte Variable_DATA[5] = { 0 }, BIT_serie = 7, Index = 0;
  unsigned int timeout;
  // Conversion du numéro de broche Arduino en ports/masque binaire "bas niveau".
  // Utiliser les registres du microcontrôleur est bien plus rapide que digitalWrite().
  uint8_t bit = digitalPinToBitMask(Broche_DATA);
  uint8_t port = digitalPinToPort(Broche_DATA);
  volatile uint8_t *ddr = portModeRegister(port);   // Registre MODE (INPUT / OUTPUT)
  volatile uint8_t *out = portOutputRegister(port); // Registre OUT (Écriture)
  volatile uint8_t *in = portInputRegister(port);   // Registre IN (Lecture)
  // >>> Réveil du capteur <<<
  *ddr |= bit; // OUTPUT.
  *out &= ~bit; // LOW.
  delay(18); // Temps d'attente à LOW engendrant le réveil du capteur.
  *out |= bit; // HIGH
  delayMicroseconds(40);
  *ddr &= ~bit; // INPUT
  // >>> Attente de la réponse du capteur <<<
  timeout = 0;
  while(!(*in & bit)) // Attente d'un état LOW.
  if (++timeout == 10000)
  return DHT11_TIMEOUT_ERROR;
  timeout = 0;
  while(*in & bit) // Attente d'un état HIGH.
  if (++timeout == 10000)
  return DHT11_TIMEOUT_ERROR;
  // Lecture des 40 bits de données du capteur.
  for (byte i = 0; i < 40; ++i) {
    // Attente d'un état LOW.
    timeout = 0;
    while(!(*in & bit))
    if (++timeout == 10000)
    return DHT11_TIMEOUT_ERROR;
    // Mise en mémoire du "temps courant".
    unsigned long t = micros();
    // Attente d'un état HIGH.
    timeout = 0;    while(*in & bit)
    if (++timeout == 10000)
    return DHT11_TIMEOUT_ERROR;
    // Si le "delta Temps" est supérieur à 40µS c'est un "1", sinon c'est un "0".
    if ((micros() - t) > 40)
    Variable_DATA[Index] |= (1 << BIT_serie); // "1"
    // Le tableau Variable_DATA[] est initialisé à "0" par défaut si le compteur de bits atteint zéro.
    if (BIT_serie-- == 0) {
      BIT_serie = 7; // On passe à l'octet suivant.
      ++Index;}}
  /* Format des données :
   * [0] = Humidité en %
   * [1] = zéro
   * [2] = Température en degrés Celsius
   * [3] = zéro
   * [4] = checksum (humidité + température)   */
  *Humidite = Variable_DATA[0];
  *Temperature = Variable_DATA[2];
  // >>> Vérifie la totalisation de contrôle. (Checksum) <<<
  if (Variable_DATA[4] != (Variable_DATA[0] + Variable_DATA[2]))
  return DHT11_CHECKSUM_ERROR;
  // Erreur de checksum.
  else    return DHT11_SUCCESS;} // Pas d'erreur.
//----------------------------------------------------------------------------------------------
 
void Analyseur_Syntaxique() {
  num_ERR = 0;
  // Extraction du nombre jusqu'à trouver un caractère non chiffre.
  // L'extraction commence au début de Chaine_Memorisee.
  NOMBRE = Chaine_Memorisee.toInt();
  Chaine_Memorisee = "";} // Vider le tampon pour une nouvelle capture.  
  
void Attendre_une_chaine_sur_le_Maitre() {// Cette procédure attend une chaine sur la ligne USB.
  Sentinelle_detectee = false;  // Sentinelle non reçue.
  Compteur_caracteres = 0;
  while (!Sentinelle_detectee) {
    if (Serial.available()) {Caractere = Serial.read(); // Puiser un caractère dans le FIFO.
    // Puis concaténer ce caractère y compris la sentinelle :
    if (Compteur_caracteres < LGR_chaine) 
       {Chaine_Memorisee = Chaine_Memorisee + Caractere; Compteur_caracteres++;}
    // Si sentinelle détectée, informer la boucle de réception :
    Sentinelle_detectee = (Caractere == '*');}}
  Repondre = true;}
  
 void BIP() {
  pinMode(BUZZER, OUTPUT); tone(BUZZER,3000); delay(100); noTone(BUZZER); pinMode(BUZZER, INPUT);}

