/* Programme Démonstrateur pour animer la Sonde JEKERT.
   Directement dérivé de P22H.
   Programme à associer avec P21J_Demonstrateur_Raquette.ino
   
   UTILISATION DU PROGRAMME :
 ----------------------------
 1) Établir la liaison TX/RX avec l'inverseur,
 2) Faire un RESET sur la Raquette,
 3) Valider l'éditeur de P22B_Demonstrateur_Sonde.ino sur le P.C.
 4) Activer le Moniteur de l'IDE. (Le P.C. affichera les ACR de la sonde, donc les n°ERR et la version.
 5) Cliquer à convenance sur toutes les touches du clavier.
    Si ERR ou version du PGM un BIP est déclenché.
 6) Tourner à convenance le codeur incrémental.
   
 Ce que fait ce démonstrateur :
 • Il indexe avec 85Pn* le PGM préservé en EEPROM.
 • Il retourne sur TX le n° du PGM pointé en EEPROM.
 • Efface le PGM indexé en EEPROM.
 • Emule le mode apprentissage.
 • Réalise le listage du PGM indexé en EEPROM.
 • Permet d'effacer le dernier code d'un PGM appris.
 • Exécuter le programme indexé en EEPROM.
 • Editer le programme indexé en EEPROM. (Ajouter des codes.)
 • Déplacer un programme en EEPROM.
 • Code pour générer une temporisation d'une seconde.
 
                                                             PGM              RAM dynamique 
Démonstrateur avant modifications :                    27406 Octets (89%)    855 Octets (41%)
Indexe, efface, et retourne le n° du PGM indexé :      27672 Octets (90%)    855 Octets (41%)   +226 octets
E13 si on tente d'apprendre sur un PGM non effacé :    27688 Octets (90%)    855 Octets (41%)    +16 octets
Emule le mode apprentissage :                          27896 Octets (90%)    858 Octets (41%)   +208 octets
Réalise le listage du PGM indexé en EEPROM :           28042 Octets (91%)    859 Octets (41%)   +146 octets
Effacer le dernier code d'un PGM appris :              28096 Octets (91%)    859 Octets (41%)    +54 octets
Executer le programme indexé en EEPROM :               28284 Octets (91%)    859 Octets (41%)   +188 octets
Editer le programme indexé en EEPROM :                 28330 Octets (92%)    859 Octets (41%)    +46 octets
Déplacer un programme en EEPROM :                      28522 Octets (92%)    860 Octets (41%)   +192 octets
Générer une temporisation d'une seconde :              28530 Octets (92%)    860 Octets (41%)     +8 octets
*/
   
//--------------------- Déclarations pour le matériel -----------------------
#include <I2Cdev.h> // Pour le MPU6050 : Directive obligatoirement placée ici.
#include <avr/eeprom.h> // Bibliothèque intégrée à l'IDE qui gère les entiers ,les réels ...
#include <Wire.h>
//#include <I2Cdev.h> // Pour le MPU6050 : Directive obligatoirement placée ici.
#include <Adafruit_PWMServoDriver.h>
Adafruit_PWMServoDriver pwm = Adafruit_PWMServoDriver();
// Déclarations pour le capteur d'humidité DHT11.
#include <DHT11.h>
#define Lecture_Adresse 0x41  //"A" en hexadécimal correspond à une commande pour le capteur Magnétique.
//--------------- Déclarations pour le module inertiel MPU6050 --------------
// Inclusion d'une bibliothèque permettant de communiquer plus facilement sur l'interface I2C.
#include <MPU6050_6Axis_MotionApps20.h>
MPU6050 Module_Inertiel;
uint16_t Taille_attendue; // Taille du paquet DMP attendu. (Par défaut : 42 octets.)
uint16_t NB_octets;       // Compte de tous les octets actuellement dans la pile FIFO.
uint8_t PILE_FIFO[64];    // Tampon de stockage de la pile FIFO.
Quaternion Q;           // [w, x, y, z]  Quaternion.
VectorFloat Vecteur_gravitation;    // [x, y, z]     Vecteur de pesanteur.
//--------------- Définition des broches d'Entrées Analogiques --------------
#define Entree_Luxmetre 6 // Entrée analogique A6 utilisée pour le luxmètre.
#define Entree_Alim_moteurs 7 // Entrée analogique A7 utilisée pour mesurer la tension de puissance.
//--------------- Définition des broches d'Entrées BINAIRES -----------------
#define Bouclier_au_sol 3 // Entrée de détection du blouclier en contact avec le sol.
//------------------ Sorties sur broches "Analogiques" ----------------------
#define LED_Mtrs_Rapides 14 // Entrée analogique A0 utilisée comme Sortie Binaire.
#define LED_Torsion_active 16 // Entrée analogique A2 utilisée comme Sortie Binaire.
#define LED_Loop_active 17 // Entrée analogique A3 utilisée comme Sortie Binaire.
//-------------- Définition des broches de sorties Binaires -----------------
#define BUZZER 2 // Le bruiteur est piloté par la sortie binaire D2.
#define Disjoncteur 4 // Masse des LEDs pour le mode Sommeil : Sortie binaire D4.
#define LED_Moteurs_OFF 7 // Pilotage LED témoin Moteurs figés.
#define Debordement_positif 8 // Pilotage LED de débordement positif du codeur rotatif.
#define LED_LASER 9 // Pilotage LED Mode LASER.
#define Broche_Com_DH11 11 // Ligne de dialogue avec le capteur d'humidité DHT11.
#define TX_Pulse_sonard 12 // TX capteur ultrasons sur la sortie D12.
#define Echo 13 // RX capteur ultrasons sur la sortie D12.
//-------------- Broches de sorties Binaires utilisées en PWM ---------------
#define Pilotage_eclairage 5 // Pilotage des deux LEDs phares de la sonde : Sortie binaire D5.
#define LASER 6 // Sortie binaire D6 pour piloter le LASER en PWM.
#define Repetiteur_CAN 10 // Sortie D10 pour piloter la LED répétiteur di CAN.
//------------------------ Constantes du programme --------------------------
#define Version_PGM "> P22-J <" // @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
#define Calibre_U_moteurs 6.67 // Calibrage du diviseur potentiométrique d'entrée. @@@@@@@@@@@@@@@@@@@@@@@@@@@
#define LGR_chaine 4 // Longueur maximale de la chaine de dialogue : Quatre en réception.
#define LGR_ACR 13 // Longueur maximale de la chaine d'ACR sur TX. @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
// Codes d'erreurs de la fonction readDHT11() :
enum {
  DHT11_SUCCESS,          // 0 : Pas d'erreur.
  DHT11_TIMEOUT_ERROR,    // 1 : Temps d'attente dépassé.
  DHT11_CHECKSUM_ERROR};  // 2 : Données reçues erronées.

//------------------------- Variables du programme --------------------------
// Note : Les déclarations sont classées par type et par ordre alphabétique croissant.
boolean Apprentissage; // Mode apprentissage en cours.
boolean Code_Valide; // Valide pour être mémorisé dans les programmes d'apprentissage.
boolean Initialiser_le_gyroscope;
boolean JEKERT_hiberne;
boolean LASER_actif; // Allumer ou éteindre le LASER.
boolean Moteurs_ON; // Si false tous les moteurs sont figés.
boolean Pb_sur_MPU6050;
boolean Phares_actifs; // Allumer ou éteindre les phares.
boolean Rapide; // Les mouvements coordonnés seront rapides.
boolean Repondre;
boolean Sentinelle_detectee;  // Délimiteur de fin détecté.
boolean Stabilisation_gyroscopique;
boolean Vertical;

char Caractere;

byte Articulation;
byte Cible_Codeur_incremental; // 0 : Personne 1 : Gradateur Phares, 2 : Gradateur LASER,
                               // 3 : Pointage LASER,    4 : Torsion,   5 : Mode_MANUEL.
byte Code_Consigne; // Contiendra la transformée numérique de la chaine reçue sur RX.
byte Compteur_caracteres;
byte Compteur_de_Code_PGM;
byte Debattement; // Pas angulaires en pointage LASER.
byte Distance; // Utilisée également pour la gestion de l'EEPROM.
byte Humidite; // Capteur d'humidité DHT11.
byte Jambe;
byte LSB; // Utilisé par le capteur magnétique.
byte MSB; // Utilisé par le capteur magnétique.
byte Niveau_eclairage_LASER; // Pilotage analogique du LASER.
byte Niveau_eclairage_Phares; // Pilotage analogique des deux phares.
byte num_ERR;
byte Num_Moteur;
byte Num_Sortie;
byte Num_PGM;
byte PGM_SOURCE;
byte Temperature; // Capteur d'humidité DHT11.
byte Valeur_Codeur_incremental;

int Adresse_du_HMC6352 = 0x42; // Écrire à cette adresse. (Capteur magnétique.)
int Angle; // Int pour pouvoir faire varier entre -90 et +90 et compatible string;
           // ATTENTION : Angle est également utilisé par l'écriture EEPROM en mode en apprentissage.
int CAN, Numerisation;
int Consigne_stabilisation;
int Lecture_Echo;
int Lumiere_ambiante;
int Pointage_Vertical;
int Pointage_Horizontal;
int PTR_Apprentissage; // Pointe en permanence le premier Octet libre du Programmes appris en EEPROM.
int PTR_Configurations; // Pointeur des octets de configuration situées en EEPROM.
int PTR_SAV_Octet; // Sauvegarder des Octets autre que PGM en EEPROM.
int Tangage, Roulis, Lacet, Cap_du_calage_gyroscope, Calage_gyro_initial;

unsigned long Ancienne_valeur_Temps_Ecoule_depuis_RESET;

float CAP_construit; // Utilisé par le capteur magnétique.
float LTR[3];           // Lacet/Tangage/Roulis.

int Configuration_actuelle[12];
int Configuration_voulue[12];
int Increment[12];

String Chaine_Memorisee; // Tampon mémoire pour la chaine entrante.
String Texte_ACR; // Tampon mémoire pour la chaine sortante.

void setup()  {
  //--------------- Mise en service de la centrale gravitationnelle ----------------
  Initialiser_le_MPU6050();
  //--------------------------------------------------------------------------------
  Serial.begin(115200);
  Chaine_Memorisee.reserve(LGR_chaine); // Reserve LGR_chaine octets pour la chaine.
  Texte_ACR.reserve(LGR_ACR); // Reserve LGR_chaine octets pour la chaine.
  //-------------------------- Initialiser des variables ---------------------------
  Valeur_Codeur_incremental = 1; Num_PGM = 1; PGM_SOURCE = 1;
  //------------------------- Multiplexeur avant les LEDs --------------------------
  pwm.begin(); // Initialise le module multiplexeur  
  pwm.setPWMFreq(50);  // Fréquence de la PWM : En standard 50Hz pour des servomoteurs.
  //--------------------------- Initialiser les sorties ----------------------------
  Phares_et_LASER_a_127(); Valeur_Codeur_incremental = 1;
  pinMode(Disjoncteur,OUTPUT); digitalWrite(Disjoncteur,HIGH); // Transistor saturé. (Masse active.)
  pinMode(LED_Loop_active,OUTPUT); pinMode(LED_Moteurs_OFF,OUTPUT); digitalWrite(LED_Moteurs_OFF,LOW);
  pinMode(LED_Torsion_active,OUTPUT); pinMode(Bouclier_au_sol,INPUT);
  pinMode(LED_Mtrs_Rapides,OUTPUT); digitalWrite(LED_Mtrs_Rapides,LOW); pinMode(Echo,INPUT);
  pinMode(TX_Pulse_sonard,OUTPUT); digitalWrite(TX_Pulse_sonard,LOW); // TX HC-SR04 ultrasons.
  pinMode(Debordement_positif,OUTPUT); Eteindre_debordements();
  pinMode(LED_LASER,OUTPUT);
  pwm.setPWM(14,0, 0); // Éteindre la LED APPRENTISSAGE.
  //--------------------- Confirmer une Sonde en hibernation ----------------------- 
  Moteurs_ON = false; CONFIGURER(192); // Impératif pour initialiser la table des positions actuelles. (VEILLE.)
  Passer_en_hibernation(); // Passe Apprentissage à false; donc pas besoin dans void setup(). 
  //--------------------------- Initialiser les booléens ---------------------------
  // Passer_en_hibernation() positionne plusieurs booléens donc placé avant ce chapitre logiciel.
  Rapide = false; // Car forcé à true par hibernation pour éteindre la LED rouge.
  Repondre = false; Cible_Codeur_incremental = 0; // Pas de cible pour (40) et (41);
  //--------------------------------------------------------------------------------
  Adresse_du_HMC6352 = Adresse_du_HMC6352 >> 1; Wire.begin(); // Capteur magnétique.
  Recale_le_gyroscope();}
  //--------------------------------------------------------------------------------

void loop() {
  //--------------------- Traite une Consigne arrivée sur USB ----------------------
  if (Serial.available()) Attendre_une_chaine_sur_RX();
  //-------------------------- Traitement de la Consignee --------------------------
  if (Chaine_Memorisee.length() > 0) Traite_la_consigne();
  //-------------------- Traitement des actions de la boucle de base ---------------
  //-------- Faire clignoter la LED de la boucle de base active --------
  LSB = 100; if (Stabilisation_gyroscopique) LSB = 30;
  if (millis() > Ancienne_valeur_Temps_Ecoule_depuis_RESET) {
      Ancienne_valeur_Temps_Ecoule_depuis_RESET = millis() + LSB; // Réarme pour 0,1S ou 0,03S.
      digitalWrite(LED_Loop_active,!digitalRead(LED_Loop_active));} // Inverser l'état de la LED.
  //-------------- Traite la LED du témoin Rapide / Lents --------------
  // Placé en boucle de base pour minimiser le code et prise en compte car gestion automatique.
  if (Rapide) digitalWrite(LED_Mtrs_Rapides,HIGH); else digitalWrite(LED_Mtrs_Rapides,LOW);   
  //----------------- Traite les fonctions analogiques -----------------
  analogWrite(Repetiteur_CAN, Valeur_Codeur_incremental);
  if (Phares_actifs) Niveau_sur_les_phares();
  else Eteindre_les_phares();
  if (LASER_actif) Niveau_sur_le_LASER();
  else Couper_le_LASER();
  //---------------- Gère la stabilisation Gyroscopique ----------------
  if (Stabilisation_gyroscopique) Stabilisation_GYROSCOPIQUE();
  //--------------- Fin du traitement des actions de la boucle de base -------------
  if (Repondre) Accuse_reception();} // ACR si une consigne a été reçue.
      
void Accuse_reception() {      
  delay(50); Repondre = false; Chaine_Memorisee = ""; 
  if (num_ERR > 0) {Serial.print("E"); Serial.print(num_ERR);} 
  else Serial.print(Texte_ACR); AFF_etoile();}

void Traite_la_consigne() {
  //----------------------------- Traite l'APPRENTISSAGE ---------------------------
  if (Apprentissage) {// num_ERR forcé à 0 par l'analyseur syntaxique;
      Filtre_les_codes();
      if (Code_Valide) {
          if (Lire_un_OCTET_en_EEPROM(Pointer_debut_PGM()) == 30) 
             {num_ERR = 14; Fin_Apprentissage();}
          else {// Incrémente le nombre de codes enregistrés dans le programme indexé en EEPROM :
                Compteur_de_Code_PGM = Lire_un_OCTET_en_EEPROM(Pointer_debut_PGM()); Compteur_de_Code_PGM++;
                Ecrire_un_OCTET_en_EEPROM(Pointer_debut_PGM(), Compteur_de_Code_PGM);
                // Enregistre le Code de la Consigne dans le programme indexé en EEPROM :
                Ecrire_un_OCTET_en_EEPROM(PTR_Apprentissage, Code_Consigne);
                PTR_Apprentissage++; }}} // Pointe l'Octet libre suivant.
  //--------------------------------------------------------------------------------
  if (num_ERR != 14) {// Si Apprentissage ET mémoire du PGM saturée ignorer la consigne.
    Texte_ACR = "OK";
    switch (Code_Consigne) {
        case  1 : {Poser_au_sol(); break;} // Commence par passer Rapide à false.
        case  2 : {Configurer_stable_transversal(); break;}
        case  3 : {Avancer_un_pas(); break;}
        case  4 : {Reculer_un_pas(); break;}
        case  5 : {Tourne_a_droite(); break;}
        case  6 : {Tourne_a_gauche(); break;}
        case  7 : {Decaler_lateralement(360); break;} // Décale latéralement à Droite.
        case  8 : {Decaler_lateralement(384); break;} // Décale latéralement à Gauche.
        case  9 : {Simule_un_Tir_au_LASER(); break;} // Laisse le LASER à niveau 1 pour pointer.
        case 10 : {Hauteur_Maximale(); break;}
        case 11 : {Retour_de_hauteur_maximale(); break;}
        case 12 : {PGM_SOURCE = Num_PGM; break;} // Le PGM actuellement indexé devient la SOURCE.
        case 13 : {Transfere_SOURCE_dans_Num_PGM(); break;} // Transfère et efface le PGM SOURCE.
        case 14 : {Libere_efforts(); break;} // Libérer les efforts.
        case 15 : {CONFIGURER(48); break;} // Tour les moteurs au neutre opérationnel.
        case 16 : {Rapide = false; Configuration_Stable_Raisonnable(); break;}
        case 17 : {Config_Stable_Raisonnable_vers_VEILLE(); break;}
        case 18 : {Enregistrement_de_la_posture_Actuelle_en_EEPROM(); break;}
        case 19 : {Utilise_la_posture_sauvegardee_en_EEPROM(); break;}
        case 20 : {Apprentissage = true; pwm.setPWM(14,0, 254); Saisie_NB_Codes_dans_le_PGM();
                   PTR_Apprentissage = Pointer_debut_PGM() + Compteur_de_Code_PGM;
                   PTR_Apprentissage++; break;}              
        case 21 : {Depose_la_sonde(); break;}
        case 22 : {CONFIGURER(96); break;} // Configuration Atterrissage.
        case 23 : {Configure_pour_Decollage(); break;}
        case 24 : {Texte_ACR = Version_PGM; break;} // S16 Indique la version du programme.   
        case 25 : {digitalWrite(Disjoncteur,LOW); break;} // S1 Disjoncte les LEDs.
        case 26 : {digitalWrite(Disjoncteur,HIGH); break;} // S2 Rétablit les LEDs.
        case 27 : {Phares_actifs = true; Niveau_sur_les_phares(); break;}
        case 28 : {Phares_actifs = false; Eteindre_les_phares(); break;}
        case 29 : {LASER_actif = true; Niveau_sur_le_LASER(); break;}
        case 30 : {Eteindre_le_LASER(); break;}
        case 31 : {Moteurs_ON = false; digitalWrite(LED_Moteurs_OFF,HIGH); break;} // S7 Figer les moteurs.
        case 32 : {Moteurs_ON = true; digitalWrite(LED_Moteurs_OFF,LOW); break;} // S8 Rétablir les moteurs.
        case 33 : {Rapide = true; break;}
        case 34 : {Rapide = false; break;}
        case 35 : {Stabilisation_gyroscopique = true; break;} // S11 Début de stabilisation gyroscopique.
        case 36 : {Stabilisation_gyroscopique = false; break;} // S12 Fin de stabilisation gyroscopique.
        case 38 : {Change_cible_codeur_incremental(4); digitalWrite(LED_Torsion_active,HIGH); CAN = 512; break;}
        case 39 : {Change_cible_codeur_incremental(0); digitalWrite(LED_Torsion_active,LOW);
                   Coordonne(120); CONFIGURER(120); break;} 
        case 40 : {Faire_un_pas_sens_Positif(); break;} // Codeur rotatif Incrémenté.
        case 41 : {Faire_un_pas_sens_Negatif(); break;} // Codeur rotatif Décrémenté.
        case 42 : {Change_cible_codeur_incremental(1); break;} // S13 Le codeur incrémental agit sur le LASER.
        case 43 : {Change_cible_codeur_incremental(2); break;} // S14 " " " " " " " " " agit sur les Phares. 
        case 44 : {Construire_ACR_etat_machine(); break;}    
        case 45 : {Texte_ACR = "U moteurs = " +
                   String(analogRead(Entree_Alim_moteurs) * Calibre_U_moteurs / 1023) + " v"; break;}
        case 46 : {Texte_ACR = String(Niveau_eclairage_Phares); break;}
        case 47 : {Texte_ACR = String(Niveau_eclairage_LASER); break;}
        case 48 : {Mesurer_Hygrometrie_et_Temperature();
                   if (num_ERR == 0) Texte_ACR = "TEMPERATURE = "+ String(Temperature) + ' ';
                   else Texte_ACR = String(num_ERR); break;}
        case 49 : {Mesurer_Hygrometrie_et_Temperature();
                   if (num_ERR == 0) Texte_ACR = "HYGROMETRIE = " + String(Humidite) + " %";
                   else Texte_ACR = String(num_ERR); break;}
        case 50 : {if (JEKERT_hiberne) Texte_ACR = "OUI"; else Texte_ACR = "NON"; break;}
        case 51 : {Telemetre_ultrasons(); Texte_ACR = "DISTANCE = " + String(Distance) + " cm"; break;}
        case 52 : {Mesure_orientation_magnetique(); Texte_ACR = "CAP actuel = " + String(Numerisation); break;}
        case 53 : {Valeur_Codeur_incremental = 1; Cible_Codeur_incremental = 0;  // Cible gradateur = Personne.
                   Eteindre_debordements(); break;}
        case 54 : {Phares_et_LASER_a_127(); break;}
        case 55 : {REVEILLER_JEKERT(); Moteurs_ON = true; break;} // Sortie de l'Hibernation.
                   //  REVEILLER_JEKERT Force Rapide à false. (Laisser Moteurs_ON = true ici.)
        case 56 : {Mesure_les_valeurs_de_gravitation();
                   Texte_ACR = "Tang. " + String(Tangage) + " Roulis " + String(Roulis); break;}
        case 57 : {Mesure_les_valeurs_de_gravitation();
                   Texte_ACR = "Calage GYRO. = " + String(Cap_du_calage_gyroscope); break;}
        case 58 : {Initialiser_le_MPU6050(); if (Pb_sur_MPU6050) num_ERR = 8; break;} // Retourne E8* ou OK*.
        case 59 : {Recale_le_gyroscope(); break;} // Recale le gyroscope sur le compas magnétique.
        case 60 : {Mesure_les_valeurs_de_gravitation();
                   Texte_ACR = "Ecart route = " + String(Lacet-Calage_gyro_initial);break;}
        case 61 : {Passer_en_hibernation(); break;}
        case 62 : {Vertical = true; Pointage_Horizontal = 305; Pointage_Vertical = 312;
                   Posture_LASER(); digitalWrite(LED_LASER,HIGH); Debattement = 12; // Eteind le LASER.
                   Change_cible_codeur_incremental(3); break;}
        case 63 : {FIN_Posture_LASER(); digitalWrite(LED_LASER,LOW); Libere_efforts(); break;}
        case 64 : {Vertical = true; break;}
        case 65 : {Vertical = false; break;}
        case 66 : {if (Debattement == 12) Debattement = 2; else Debattement = 12; break;}
        case 67 : {Enregistrer_Spectroscope_Chromatique(); break;}
        case 68 : {Envoyer_le_Spectroscope_Chromatique(); break;}
        case 69 : {Enregistrer_un_Balayage_telemetrique(); break;}
        case 70 : {Envoyer_le_Balayage_telemetrique(); break;}
        case 71 : {if (Lire_un_OCTET_en_EEPROM(Pointer_debut_PGM()) != 0) num_ERR = 13;
                   else {Apprentissage = true; pwm.setPWM(14,0, 254);
                         PTR_Apprentissage = Pointer_debut_PGM(); PTR_Apprentissage++;} break;}
        case 72 : {Fin_Apprentissage(); break;}
        case 73 : {Cible_Codeur_incremental = 5; Debattement = 12; break;} 
        case 74 : {Cible_Codeur_incremental = 0; break;}
        case 75 : {Jambe = 0; Calcule_num_sortie(); break;} // Jambe A.
        case 76 : {Jambe = 3; Calcule_num_sortie(); break;} // Jambe B.
        case 77 : {Jambe = 6; Calcule_num_sortie(); break;} // Jambe C.
        case 78 : {Jambe = 9; Calcule_num_sortie(); break;} // Jambe B.
        case 79 : {Articulation = 0;  Calcule_num_sortie(); break;} // Hanche.
        case 80 : {Articulation = 1;  Calcule_num_sortie(); break;} // Genou.
        case 81 : {Articulation = 2;  Calcule_num_sortie(); break;} // Griffe.
        case 82 : {Retourner_les_valeurs_de_posture(true); break;} // Retourne la posture actuelle;
        case 83 : {PTR_Configurations = 408; Charge_configuration_voulue(); 
                   Retourner_les_valeurs_de_posture(false); break;} // Retourne la posture enregistrée;
        case 84 : {Texte_ACR = String(Num_PGM); break;} // Indique le n° du PGM actuellement actif.
        case 85 : {Chaine_Memorisee = Chaine_Memorisee.substring(3); // Impose le n° du PGM actuellement actif.
                   Num_PGM = Chaine_Memorisee.toInt(); break;}
        case 86 : {Retourner_le_Listage_du_PGM_cible(); break;} // Liste le PGM indexé en EEPROM.
        case 87 : {Activer_le_PGM_cible(); break;} // Déclencherle déroulement du PGM indexé en EEPROM.
        case 88 : {Ecrire_un_OCTET_en_EEPROM(Pointer_debut_PGM(),0); break;} // Efface le programme Num_PGM.                
        case 89 : {Efface_le_dernier_code_appris(); break;}
        case 90 : {delay(1000); break;} // Génère une temporisation d'une seconde.
        default : num_ERR = 5;}}}

//----------------------------------- Gestion de l'APPRENTISSAGE -------------------------------
void Filtre_les_codes() {// Un "switch /case" n'est pas rentable, la suite de "if" est optimale.
  Code_Valide = true;
  if (Code_Consigne == 12) Code_Valide = false;
  if (Code_Consigne == 13) Code_Valide = false;
  if (Code_Consigne == 18) Code_Valide = false;
  if (Code_Consigne == 20) Code_Valide = false;
  if (Code_Consigne == 23) Code_Valide = false;
  if (Code_Consigne == 24) Code_Valide = false;
  if (Code_Consigne == 37) Code_Valide = false;
  if ((Code_Consigne > 43) && (Code_Consigne < 53)) Code_Valide = false;
  if (Code_Consigne == 56) Code_Valide = false;
  if (Code_Consigne == 57) Code_Valide = false;
  if (Code_Consigne == 59) Code_Valide = false;
  if (Code_Consigne == 60) Code_Valide = false;
  if (Code_Consigne == 68) Code_Valide = false;
  if (Code_Consigne == 70) Code_Valide = false;
  if (Code_Consigne == 72) Code_Valide = false;
  if ((Code_Consigne > 81) && (Code_Consigne < 90)) Code_Valide = false;}
  
void Transfere_SOURCE_dans_Num_PGM() {
  byte SAV_Num_PGM = Num_PGM;
  if (PGM_SOURCE == Num_PGM) num_ERR = 23;
  Saisie_NB_Codes_dans_le_PGM(); if (Compteur_de_Code_PGM != 0) num_ERR = 24; 
  if (num_ERR == 0) {Recopier_les_Codes(); Num_PGM = PGM_SOURCE;
                     Ecrire_un_OCTET_en_EEPROM(Pointer_debut_PGM(),0); Num_PGM = SAV_Num_PGM;}}
                     
void Recopier_les_Codes() {
  PTR_Apprentissage = Pointer_debut_PGM(); // PTR_Apprentissage pointe la destination.
  Num_PGM = PGM_SOURCE; PTR_Configurations = Pointer_debut_PGM(); // PTR_Configurations pointe la SOURCE.
  for (byte I = 0; I < 31; I++) { // Copie de 31 octets pour transférer également le nombre de codes du PGM.
       LSB = Lire_un_OCTET_en_EEPROM(PTR_Configurations); PTR_Configurations++; 
       Ecrire_un_OCTET_en_EEPROM(PTR_Apprentissage,LSB); PTR_Apprentissage++;}}
  
void Activer_le_PGM_cible() {
  Saisie_NB_Codes_dans_le_PGM(); // Positionne le pointeur de codes sur [début du PGM - 1].
  if (Compteur_de_Code_PGM == 0) num_ERR = 18;
  else {num_ERR = 0;
        while (Compteur_de_Code_PGM > 0) {
          PTR_Apprentissage++; // Passer au code suivant.
          Code_Consigne = Lire_un_OCTET_en_EEPROM(PTR_Apprentissage);
          delay(200); AFF_etoile(); // Pour barrer le code sur la grille de listage.
          Traite_la_consigne();
          if (num_ERR == 0) Compteur_de_Code_PGM--; else {num_ERR = 20; Compteur_de_Code_PGM = 0;}}}}
  
void Efface_le_dernier_code_appris() {
  Saisie_NB_Codes_dans_le_PGM();
  if (Compteur_de_Code_PGM == 0) num_ERR = 17;
  else {Compteur_de_Code_PGM--; Ecrire_un_OCTET_en_EEPROM(PTR_Apprentissage, Compteur_de_Code_PGM);}}
  
void Saisie_NB_Codes_dans_le_PGM() {
  PTR_Apprentissage = Pointer_debut_PGM(); // Pointe le Nb de codes du PGM en EEPROM.
  Compteur_de_Code_PGM = Lire_un_OCTET_en_EEPROM(PTR_Apprentissage);}
  
void Retourner_le_Listage_du_PGM_cible() {
  Saisie_NB_Codes_dans_le_PGM();
  if (Compteur_de_Code_PGM == 0) num_ERR = 15;
  else {PTR_Apprentissage++; // Poiter le premier octet du programme;
        while (Compteur_de_Code_PGM > 0) {delay(100); //70
               Serial.print(Lire_un_OCTET_en_EEPROM(PTR_Apprentissage)); AFF_etoile();
               PTR_Apprentissage++; Compteur_de_Code_PGM--;}}}
      
int Pointer_debut_PGM() {return 448 + (Num_PGM * 32);}

void Fin_Apprentissage() {Apprentissage = false; pwm.setPWM(14,0, 0);}

void Niveau_sur_les_phares() {analogWrite(Pilotage_eclairage, Niveau_eclairage_Phares);}
void Eteindre_les_phares(){analogWrite(Pilotage_eclairage, 0);}
void Niveau_sur_le_LASER() {analogWrite(LASER, Niveau_eclairage_LASER);}
void Couper_le_LASER() {analogWrite(LASER, 0);}

//----------------------------- Retourner les consignes d'une POSTURE --------------------------
void Retourner_les_valeurs_de_posture(boolean Actuelle) {// Lecture_Echo est utilisé pour économiser une variable.
  for (byte I=0; I < 12; I++) {
  if (Actuelle) Lecture_Echo = Configuration_actuelle[I]; else {Lecture_Echo = Configuration_voulue[I];}
  Texte_ACR = String(Lecture_Echo) + '*'; delay(100); Serial.print(Texte_ACR);}
  Repondre = false; Chaine_Memorisee ="";}
  
//--------------------------------- Piloter_un_moteur_en_manuel --------------------------------
void Calcule_num_sortie() {Num_Sortie = Jambe + Articulation;}

void Ajuste_la_position_du_servomoteur(boolean Augmente) {
  int Consigne;
  Consigne = Configuration_actuelle[Num_Sortie];
  if (Augmente) Consigne = Consigne + Debattement; else Consigne = Consigne - Debattement;
  Mouvoir(Num_Sortie, Consigne);}

//-------------------------- Enregistrement d'un balayage Télémétrique -------------------------
void Envoyer_le_Balayage_telemetrique() {
  byte I=1;
  PTR_SAV_Octet = 432;
  while (PTR_SAV_Octet < 464) 
        {Distance = Lire_un_OCTET_en_EEPROM(PTR_SAV_Octet); PTR_SAV_Octet++;
         if (Distance > 52) Distance = 52; // Limiter la longueur des lignes affichées.
         Texte_ACR = String(Distance) + '*'; delay(50); Serial.print(Texte_ACR);}
   Repondre = false; Chaine_Memorisee ="";}       
  
void Enregistrer_un_Balayage_telemetrique() {// Ne faire qu'une procédure n'économise pas du code objet.
  Aller_au_Debut(); Bouge_pour_Enregistrer (); Revenir_au_centre();
  Configurer_stable_transversal();}
 
void Bouge_pour_Enregistrer () {
  PTR_SAV_Octet = 432;
  for (CAN = 47; CAN < 1008; CAN = CAN + 30)
      {Bouge_pour_un_balayage_en_torsion(); Telemetre_ultrasons(); Distance = Distance >> 2; delay(30);
       Ecrire_un_OCTET_en_EEPROM(PTR_SAV_Octet, Distance); PTR_SAV_Octet++;}}
      
void Aller_au_Debut() {
  for (CAN = 512; CAN > 31; CAN = CAN - 31) Bouge_pour_un_balayage_en_torsion();}

void Revenir_au_centre() {
  for (CAN = 1008; CAN > 512; CAN = CAN - 31) Bouge_pour_un_balayage_en_torsion();}

//-------------------------- Routines pour la correction d'orientation -------------------------
void Bouge_pour_un_balayage_en_torsion() {
  Mouvoir(0, (map(CAN,47,1008,325,247)) - 47);  // Hanche de la Jambe A.
  Mouvoir(3, (map(CAN,47,1008,345,274)) + 35);  // Hanche de la Jambe B.
  Mouvoir(6, (map(CAN,47,1008,304,228)) - 44);  // Hanche de la Jambe C.
  Mouvoir(9, (map(CAN,47,1008,344,264)) + 42);} // Hanche de la Jambe D.

//-------------------------- Routines pour la spectroscopie chromatique ------------------------
void Enregistrer_Spectroscope_Chromatique() {// Lecture_Echo contient la consigne angulaire de positionnement.
     Posture_LASER();
     PTR_SAV_Octet = 464; Lecture_Echo = 211; // Positionne sur "transparent".
     for (byte I=1; I < 8; I++) {
     Mouvoir(0, Lecture_Echo); delay(200);
     Lecture_Echo = Lecture_Echo + 34; // Ajoute un "angle" pour aller à la couleur suivante.   
     Lumiere_ambiante = analogRead(Entree_Luxmetre);
     Ecrire_un_Entier_en_EEPROM(PTR_SAV_Octet, Lumiere_ambiante);}
     FIN_Posture_LASER();}
     
void Envoyer_le_Spectroscope_Chromatique() {
  PTR_SAV_Octet = 464;
  for (byte I=1; I < 8; I++) {
  Lecture_Echo = Lire_un_Entier_en_EEPROM(PTR_SAV_Octet); PTR_SAV_Octet++; PTR_SAV_Octet++;
  Texte_ACR = String(Lecture_Echo) + '*'; delay(100); Serial.print(Texte_ACR);}
  Repondre = false; Chaine_Memorisee ="";} 
      
//------------------------------- Procédure d'utilisation du LASER -----------------------------
void Eteindre_le_LASER() {LASER_actif = false; Couper_le_LASER();}

void Posture_LASER() {// Passe en Posture Stable LASER, la position stable sur trois pattes.
  Stabilisation_gyroscopique = false; // Fonction incompatible avec la posture LASER.
  Eteindre_le_LASER();
  Configurer_stable_transversal(); // Préalable pour positionner en posture LASER.
  for (byte Nb = 1; Nb < 6; Nb++) {Avance_Recule_Jambe(true, 'D'); delay(100);}
  Coordonne(216); Libere_efforts();}// Patte A en configuration de pointage.

void FIN_Posture_LASER() {
  Eteindre_le_LASER();
  CONFIGURER(240); delay(300); // Configuration LASER jambe A ramenée au stable transversal.
  for (byte Nb = 1; Nb < 6; Nb++) {Avance_Recule_Jambe(false, 'D'); delay(100);}}
 
void Pointer_plus_Haut() {
  Pointage_Vertical = Pointage_Vertical + Debattement; 
  if (Pointage_Vertical > 415) {Pointage_Vertical = 415; num_ERR = 9;}
  Mouvoir(1, Pointage_Vertical);}

void Pointer_plus_Bas() {
  Pointage_Vertical = Pointage_Vertical - Debattement;
  if (Pointage_Vertical < 215) {Pointage_Vertical = 215; num_ERR = 10;}
  Mouvoir(1, Pointage_Vertical);}

void Pointer_plus_a_Gauche() {
  Pointage_Horizontal = Pointage_Horizontal + Debattement;
  if (Pointage_Horizontal > 395) {Pointage_Horizontal = 395; num_ERR = 9;} 
  Mouvoir(0, Pointage_Horizontal);}

void Pointer_plus_a_Droite() {
  Pointage_Horizontal = Pointage_Horizontal - Debattement;
  if (Pointage_Horizontal < 220) {Pointage_Horizontal = 220; num_ERR = 10;}
  Mouvoir(0, Pointage_Horizontal);}

//------------------------------------ Procesures de service -----------------------------------
void Faire_un_pas_sens_Positif() {
  switch (Cible_Codeur_incremental) {
    case 1 : {Augmente_energie(); Niveau_eclairage_LASER = Valeur_Codeur_incremental; Niveau_sur_le_LASER(); break;}
    case 2 : {Augmente_energie(); Niveau_eclairage_Phares = Valeur_Codeur_incremental; Niveau_sur_les_phares(); break;}
    case 3 : {if (Vertical) Pointer_plus_Haut(); else Pointer_plus_a_Droite(); break;}
    case 4 : {Tordre(true); break;}
    case 5 : {Ajuste_la_position_du_servomoteur(true); break;}}
  if (num_ERR > 0) digitalWrite(Debordement_positif,HIGH); else Eteindre_debordements();}
  
void Faire_un_pas_sens_Negatif() {
  switch (Cible_Codeur_incremental) {
    case 1 : {Diminue_energie(); Niveau_eclairage_LASER = Valeur_Codeur_incremental; break;}
    case 2 : {Diminue_energie(); Niveau_eclairage_Phares = Valeur_Codeur_incremental; break;}
    case 3 : {if (Vertical) Pointer_plus_Bas(); else Pointer_plus_a_Gauche(); break;}
    case 4 : {Tordre(false); break;}
    case 5 : {Ajuste_la_position_du_servomoteur(false); break;}}
  if (num_ERR > 0) pwm.setPWM(13,0, 254); else Eteindre_debordements();}
  
//------------------------------- Traite les mouvements de Torsion -----------------------------
 void Tordre(boolean A_droite) {
   if (A_droite) CAN = CAN - 20;
   else CAN = CAN + 20;
   Bouge_pour_un_balayage_en_torsion();} 
  
//-------------------------- Gère l'enregistrement de posture en EEPROM ------------------------
void  Enregistrement_de_la_posture_Actuelle_en_EEPROM() {
      PTR_SAV_Octet = 408;
      for (byte I = 0; I < 12; I++) {
           Ecrire_un_Entier_en_EEPROM(PTR_SAV_Octet, Configuration_actuelle[I]);}}  
     
//=========================== Routines pour la Stabilisation GYROSCOPIQUE ============================
void Stabilise(boolean Plus, byte Sortie) {
  Consigne_stabilisation = Configuration_actuelle[Sortie];
  if (Plus) Consigne_stabilisation++; else Consigne_stabilisation--;
  Mouvoir(Sortie,Consigne_stabilisation);}
  
void Stabilisation_GYROSCOPIQUE() {
    Mesure_les_valeurs_de_gravitation();
    if (Roulis > 0) {
        Stabilise(true, 1); Stabilise(false, 4); Stabilise(false, 7); Stabilise(true, 10);}
    if (Roulis < 0) {
      Stabilise(false, 1); Stabilise(true, 4); Stabilise(true, 7); Stabilise(false, 10);}
    if (Tangage > 0) {
       Stabilise(true, 1); Stabilise(true, 2); Stabilise(true, 4); Stabilise(true, 5);
       Stabilise(false, 7); Stabilise(false, 8); Stabilise(false, 10); Stabilise(false, 11);}
    if (Tangage < 0) {
       Stabilise(false, 1); Stabilise(false, 2); Stabilise(false, 4); Stabilise(false, 5);
       Stabilise(true, 7); Stabilise(true, 8); Stabilise(true, 10); Stabilise(true, 11);} }
       
//--------------------------------- Capeur gravitationnel MPU6050 ------------------------------
void Mesure_les_valeurs_de_gravitation() {
  Module_Inertiel.resetFIFO(); // Reset pour vider le FIFO et s'assurer d'une transmission correcte.
  NB_octets = Module_Inertiel.getFIFOCount(); // Acquisition du compte actuel de la pile FIFO.
  // Surveillance d'overflow. (Qui ne devrait jamais se produire sauf si le code est inefficace.)
  if (NB_octets == 1024) Module_Inertiel.resetFIFO(); // Reset pour pouvoir continuer proprement. 
  else { // Attente en principe très courte des données de longueur correcte.
    while (NB_octets < Taille_attendue) {NB_octets = Module_Inertiel.getFIFOCount(); }
    // Lecture d'un paquet dans la pile FIFO.
     Module_Inertiel.getFIFOBytes(PILE_FIFO, Taille_attendue);
    // Mise à jour du compte de la pile FIFO dans le cas où il y a plus d'un paquet disponible
    // ce qui permet d'en lire immédiatement plus sans attendre une interruption.
    NB_octets -= Taille_attendue;
    // Détermination des angles d'Euler en degrés :
    Module_Inertiel.dmpGetQuaternion(&Q, PILE_FIFO);
    Module_Inertiel.dmpGetGravity(&Vecteur_gravitation, &Q);
    Module_Inertiel.dmpGetYawPitchRoll(LTR, &Q, &Vecteur_gravitation);
    Tangage = (-LTR[1] * 180/M_PI); Roulis = (LTR[2] * 180/M_PI);  Lacet = (LTR[0] * 180/M_PI); }}   
  
void Initialiser_le_MPU6050() {// Retourne la valeur du test dans le booleen Pb_sur_MPU6050;  
  Initialiser_le_gyroscope = true; Pb_sur_MPU6050 = false;
  Serial.flush(); // S'assurer que le tampon de transmission est vide.
  Wire.begin(0x20); //Initialise la liaison I2C avec un "esclave".
  // Initialisation du capteur accélérométrique et gyroscopique :
  Module_Inertiel.initialize();
  // Chargement et configuration du DMP :
  int devStatus = Module_Inertiel.dmpInitialize();
  if (devStatus == 0) {
    // Activation du module DMP6050.
    Module_Inertiel.setDMPEnabled(true); // Activation du DMP.
    // Acquisition de la taille de paquet DMP pour comparaison ultérieure
    Taille_attendue = Module_Inertiel.dmpGetFIFOPacketSize(); }
    else Pb_sur_MPU6050 = true; // Pour pouvoir interroger suite à un RESET.
  if (Module_Inertiel.testConnection()) // Test de la connexion.
      Module_Inertiel.resetFIFO(); // Reset pour vider le FIFO et s'assurer d'une transmission correcte.
  else Pb_sur_MPU6050 = true;}
  
void Recale_le_gyroscope() {
     Calage_gyro_initial = Lacet; Mesure_orientation_magnetique(); Cap_du_calage_gyroscope = Numerisation;}
   
//------------------------------------ Routines pour MOUVOIR -----------------------------------
void Mouvoir(byte Sortie, int Consigne) {
  if (Moteurs_ON) {pwm.setPWM(Sortie,0, Consigne); Configuration_actuelle[Sortie] = Consigne;}}

void Bouge_un_membre(byte Num_Sortie, int Consigne) {
  Mouvoir(Num_Sortie, Consigne); Configuration_actuelle[Num_Sortie] = Consigne;}
  
void Libere_efforts() {
  Lever_Baisser_delai_100(true,'A'); Lever_Baisser_delai_100(false,'A'); 
  Lever_Baisser_delai_100(true,'C'); Lever_Baisser_delai_100(false,'C'); 
  Lever_Baisser_delai_100(true,'B'); Lever_Baisser_delai_100(false,'B'); 
  Lever_Baisser_delai_100(true,'D'); Lever_Baisser_delai_100(false,'D');}
  
void Lever_Baisser_delai_100(boolean Leve, char Membre) {Lever_ou_Baisser_Jambe(Leve, Membre); delay(100);}

void Lever_ou_Baisser_Jambe(boolean Lever, char Jambe) {
  Caractere = Jambe; Transpose_char_vers_Sortie();
  Angle = 28; if ((Jambe == 'B') || (Jambe == 'D')) Angle = -28;
  if (Lever) {Bouge_un_membre(Num_Sortie + 1, Configuration_actuelle[Num_Sortie + 1] + Angle);}
  else {Bouge_un_membre(Num_Sortie + 1, Configuration_actuelle[Num_Sortie + 1] - Angle);}
  Angle = 47; if ((Jambe == 'B') || (Jambe == 'D')) Angle = -47;
  if (Lever) {Bouge_un_membre(Num_Sortie + 2, Configuration_actuelle[Num_Sortie + 2] + Angle);}
  else {Bouge_un_membre(Num_Sortie + 2, Configuration_actuelle[Num_Sortie + 2] - Angle);}}
  
void Transpose_char_vers_Sortie() {
  switch (Caractere) {
    case'A' : { Num_Sortie = 0; break;}
    case'B' : { Num_Sortie = 3; break;}
    case'C' : { Num_Sortie = 6; break;}
    case'D' : Num_Sortie = 9;} }
    
void Reculer_un_pas() {
  Recule_paire(false); Recule_paire(true); Coordonne(168); 
  Recule_paire(false); Recule_paire(true); Configurer_stable_transversal();}

void Avancer_un_pas() { // Anticollision par télémètre à ultrasons.
  Telemetre_ultrasons();
  if (Distance > 8) {Avance_paire(false); Avance_paire(true); Coordonne(264);
                     Avance_paire(false); Avance_paire(true); Configurer_stable_transversal();}
  else {num_ERR = 6;} }
  
void Decaler_lateralement(int Posture) {
  Poser_au_sol(); CONFIGURER(96); delay(500); Coordonne(Posture);
  REVEILLER_JEKERT();} // En réalité relève la sonde et la stabilise.
  
void Tourne_a_droite() {
   Avance_paire(true); Recule_paire(false); Coordonne(312); // Fait tourner.
   Recule_paire(false); Avance_paire(true); Configurer_stable_transversal(); }
   
void Tourne_a_gauche() {
   Avance_paire(false); Recule_paire(true); Coordonne(336); // Fait tourner.
   Avance_paire(false); Recule_paire(true); Configurer_stable_transversal(); }
 
void Recule_paire(boolean Babord)
  {if (Babord) {Avance_Recule_Jambe (false , 'C'); delay(200);
       Avance_Recule_Jambe (false , 'D'); delay(200);}
   else {Avance_Recule_Jambe (false , 'A'); delay(200);
         Avance_Recule_Jambe (false , 'B'); delay(200);} } 

void Avance_paire(boolean Babord)
  {if (Babord) {Avance_Recule_Jambe (true , 'C'); delay(200);
      Avance_Recule_Jambe (true , 'D'); delay(200);}
   else {Avance_Recule_Jambe (true , 'A'); delay(200);
         Avance_Recule_Jambe (true , 'B'); delay(200);} }
        
void Avance_Recule_Jambe (boolean Avancer, char Jambe) {
  int Pas;
  if (Avancer) {Pas = 35; if ((Jambe == 'C') || (Jambe == 'D')) Pas = -35;}
  else {Pas = -35; if ((Jambe == 'C') || (Jambe == 'D')) Pas = +35;}
  Lever_ou_Baisser_Jambe(true, Jambe); delay(50);
  Bouge_un_membre(Num_Sortie, Configuration_actuelle[Num_Sortie] + Pas); delay(50);
  Lever_ou_Baisser_Jambe(false, Jambe);}
      
//--------------------------------- Routines pour les POSTURES ---------------------------------
void Poser_au_sol() {Rapide = false; Coordonne(192);}

void Configurer_stable_transversal() {Coordonne(120); CONFIGURER(120);}

void REVEILLER_JEKERT() {digitalWrite(Disjoncteur,HIGH); JEKERT_hiberne = false;
                  Rapide = false; Configurer_stable_transversal(); Libere_efforts();}
    
void Passer_en_hibernation() // l'ordre des actions est important.
  {Fin_Apprentissage(); Eteindre_le_LASER(); Niveau_eclairage_LASER = 1; 
   Moteurs_ON = true; Poser_au_sol(); // Passe Rapide à false.
   Moteurs_ON = false; digitalWrite(Disjoncteur,LOW); Phares_actifs = false;
   Change_cible_codeur_incremental(0); Stabilisation_gyroscopique = false;
   Niveau_eclairage_Phares = 1; JEKERT_hiberne = true;
   digitalWrite(LED_Torsion_active,LOW);}  

void Retour_de_hauteur_maximale() { Coordonne(288); REVEILLER_JEKERT();}
  
void Configuration_Stable_Raisonnable() {
  Coordonne(0); CONFIGURER(0); Libere_efforts();}
  
void Config_Stable_Raisonnable_vers_VEILLE() {
  Coordonne(120); Coordonne(192); }
   
void Hauteur_Maximale() {Coordonne(288); delay(300); Coordonne(144);}

void Depose_la_sonde() {
  if (Contact_avec_le_sol()) {
      CONFIGURER(48); // Tous les moteurs au neutre opérationnel.
      delay(1000); Poser_au_sol(); delay(1000); REVEILLER_JEKERT();}  
  else num_ERR = 7;}
  
void Configure_pour_Decollage() {
  if (Contact_avec_le_sol()) {Coordonne(72); CONFIGURER(72);}
  else num_ERR = 7;}
  
boolean Contact_avec_le_sol() {return digitalRead(Bouclier_au_sol);}

void Utilise_la_posture_sauvegardee_en_EEPROM() {Coordonne(408); CONFIGURER(408);}

void Charge_configuration_voulue() {
  for (byte I=0; I < 12; I++)
  {Configuration_voulue[I] = Lire_un_Entier_en_EEPROM(PTR_Configurations);
   PTR_Configurations++; PTR_Configurations++;}}

void CONFIGURER(int PTR_TABLE) {
  for (Num_Sortie = 0; Num_Sortie < 12; Num_Sortie++) {
       Bouge_un_membre(Num_Sortie, Lire_un_Entier_en_EEPROM(PTR_TABLE)); PTR_TABLE++; PTR_TABLE++;} }
 
void Coordonne(int Configuration_desiree) {
  // Passe de configuration actuelle en configuration désirée en 8 étapes.
  PTR_Configurations = Configuration_desiree; Charge_configuration_voulue();
  // Passe de configuration actuelle en configuration désirée en 8 étapes.
  for (byte I = 0; I < 12; I++)
       Increment[I] = (Configuration_voulue[I] - Configuration_actuelle[I]) >> 3;
  for (byte I = 1; I < 9; I++)
      {for (byte sortie = 0; sortie < 12; sortie++)
      Bouge_un_membre(sortie, Configuration_actuelle[sortie] + Increment[sortie]);
      if (Rapide) delay(50); else delay(200);} }      

//-------------------- Routines pour lire ou écrire des données en EEPROM  ---------------------
byte Lire_un_OCTET_en_EEPROM(int ADRESSE) {return eeprom_read_byte((unsigned char *) ADRESSE);}

void Ecrire_un_OCTET_en_EEPROM(int ADRESSE, byte OCTET) { // Attention : Ne modifie pas le pointeur.
   eeprom_write_byte((unsigned char *) ADRESSE, OCTET);}
   
int Lire_un_Entier_en_EEPROM(int ADRESSE) {return eeprom_read_word((unsigned int *) ADRESSE);}

void Ecrire_un_Entier_en_EEPROM(int ADRESSE, int MOT16Bits) { // Pointe l'octet libre suivant.
   eeprom_write_word((unsigned int *) ADRESSE, MOT16Bits); PTR_SAV_Octet++; PTR_SAV_Octet++;}
      
//--------------------------------- Capteur magnétique HMC6352 ---------------------------------
void Mesure_orientation_magnetique() {// Saisie de la valeur du cap magnétique. 
  Wire.beginTransmission(Adresse_du_HMC6352);
  Wire.write(Lecture_Adresse); // Commande de lecture.
  Wire.endTransmission();
  delay(6); // le HMC6352 a besoin d'un délai de 6ms quand il reçoit une commande.
  Wire.requestFrom(Adresse_du_HMC6352, 2);
  // Saisie des deux octets de donnée "MSB" et "LSB".
  MSB = Wire.read(); LSB = Wire.read();
  // ========== Traite la donnée et calcule l'orientation. ==========
  // La donnée issue de deux octets est concaténée, "MSB" étant décalé huit fois à gauche.
  CAP_construit = (MSB << 8) + LSB;
  // =========== Correction automatique du compas de route ==========
  // La donnée est fournie en degrés et dixièmes de degrés d'où la division par 10.
  Numerisation = int(CAP_construit / 10); // Mesure en degrés.
  if ((Numerisation >= 0) && (Numerisation < 86)) LSB = 1;     // De 0° à 85°.
  if ((Numerisation > 85) && (Numerisation < 144)) LSB = 2;    // De 86° à 143°.
  if ((Numerisation > 143) && (Numerisation < 308)) LSB = 3;   // De 144° à 307°.
  if ((Numerisation > 307) && (Numerisation <= 359)) LSB = 4;  // De 308° à 359°.
  switch (LSB) {
      case 1 : {CAN = map(Numerisation,0,85,+30,-40); Numerisation = Numerisation + CAN; break;}
      case 2 : {Numerisation = Numerisation - 40; break;}
      case 3 : {CAN = map(Numerisation,144,307,-40,+28); Numerisation = Numerisation + CAN; break;}
      case 4 : Numerisation = Numerisation + 29;}
if (Numerisation > 359) Numerisation = Numerisation -360; }

//------------------------------------ Routines de servitude -----------------------------------
void Simule_un_Tir_au_LASER() {
  for (byte N = 0; N < 35; N++)
       {analogWrite(LASER,254); delay(25); analogWrite(LASER,1); delay(50);}}
       
void Phares_et_LASER_a_127() {Niveau_eclairage_LASER = 127; Niveau_eclairage_Phares = 127;}

void Construire_ACR_etat_machine() {
 if (digitalRead(Disjoncteur)) Texte_ACR = "0"; else Texte_ACR = "1";
 Ajoute_booleen(Moteurs_ON); Ajoute_booleen(Rapide); Ajoute_booleen(Stabilisation_gyroscopique);
 if (digitalRead(Bouclier_au_sol)) Texte_ACR = Texte_ACR + "1"; else Texte_ACR = Texte_ACR + "0";
 Ajoute_booleen(Phares_actifs); Ajoute_booleen(LASER_actif); Ajoute_booleen(Apprentissage);}

void Ajoute_booleen(boolean ETAT) {
 if (ETAT) Texte_ACR = Texte_ACR + "1"; else Texte_ACR = Texte_ACR + "0";} 
    
void Eteindre_debordements() {pwm.setPWM(13,0, 0); digitalWrite(Debordement_positif,LOW);}

void Change_cible_codeur_incremental(byte CIBLE) {
  Cible_Codeur_incremental = CIBLE;
  switch (CIBLE) {
    case 1 : {Valeur_Codeur_incremental = Niveau_eclairage_LASER; break;} // Le codeur agit sur le LASER.
    case 2 : {Valeur_Codeur_incremental = Niveau_eclairage_Phares; break;}}} // Le codeur agit sur les Phares.

void Augmente_energie() {
  if (Valeur_Codeur_incremental < 236) Valeur_Codeur_incremental = Valeur_Codeur_incremental + 18;
  else {num_ERR = 1; Valeur_Codeur_incremental = 254;}} // Ne pas dépasser 254.
  
void Diminue_energie() {
 if (Valeur_Codeur_incremental > 18) Valeur_Codeur_incremental = Valeur_Codeur_incremental - 18;
 else {num_ERR = 2; Valeur_Codeur_incremental = 1;}} // Ne pas descendre en dessous de 1.
 
//--------------------------- Télémétrie à capteur Ultrasons HC-SR04 ---------------------------
void Telemetre_ultrasons() {
  // Déclencher une mesure :
  digitalWrite(TX_Pulse_sonard, HIGH); delayMicroseconds(10); digitalWrite(TX_Pulse_sonard, LOW);
  // Analyser le retour sur la ligne de réception :
  Distance = pulseIn(Echo, HIGH) / 58;}
 
//----------------------------- Routines du capteur d'Humidité DHT11 ---------------------------
/* >>> Lecture de la température et lu taux d'humidité mesurés par un capteur DHT11 <<<
  * Broche_DATA : Broche sur laquelle est câblée la sortie du capteur.
  * Temperature : Pointeur vers la variable stockant la température.
  * Humidite : Pointeur vers la variable stockant le taux d'humidité.
  * Retourne DHT11_SUCCESS si aucune erreur, DHT11_TIMEOUT_ERROR en cas de dépassement,
    ou DHT11_CHECKSUM_ERROR en cas d'erreur de checksum */

void Mesurer_Hygrometrie_et_Temperature() {
  // >>> Lecture de la température et de l'humidité + gestion des erreurs <<<
  switch (readDHT11(Broche_Com_DH11, &Temperature, &Humidite)) {
    case DHT11_SUCCESS: break; //Ne rien faire si OK.          
    case DHT11_TIMEOUT_ERROR: {num_ERR = 3; break;}
    case DHT11_CHECKSUM_ERROR: num_ERR = 4;}} // Fin du case DHT11_SUCCESS:
   
byte readDHT11(byte Broche_DATA, byte* Temperature, byte* Humidite) {
    // Variable_DATA[] : Tampon contenant les données du cpateur.
    // BIT_serie : Compteur permettant de savoir quel bit est reçu. (bitwise)
    // Index : Compteur permettant de savoir quel octet est reçu. (bitwise)
    // timeout : Compteur pour le délai en dépassement.
  byte Variable_DATA[5] = { 0 }, BIT_serie = 7, Index = 0;
  unsigned int timeout;
  // Conversion du numéro de broche Arduino en ports/masque binaire "bas niveau".
  // Utiliser les registres du microcontrôleur est bien plus rapide que digitalWrite().
  uint8_t bit = digitalPinToBitMask(Broche_DATA);
  uint8_t port = digitalPinToPort(Broche_DATA);
  volatile uint8_t *ddr = portModeRegister(port);   // Registre MODE (INPUT / OUTPUT)
  volatile uint8_t *out = portOutputRegister(port); // Registre OUT (Écriture)
  volatile uint8_t *in = portInputRegister(port);   // Registre IN (Lecture)
  // >>> Réveil du capteur <<<
  *ddr |= bit; // OUTPUT.
  *out &= ~bit; // LOW.
  delay(18); // Temps d'attente à LOW engendrant le réveil du capteur.
  *out |= bit; // HIGH
  delayMicroseconds(40);
  *ddr &= ~bit; // INPUT
  // >>> Attente de la réponse du capteur <<<
  timeout = 0;
  while(!(*in & bit)) // Attente d'un état LOW.
  if (++timeout == 10000)
  return DHT11_TIMEOUT_ERROR;
  timeout = 0;
  while(*in & bit) // Attente d'un état HIGH.
  if (++timeout == 10000)
  return DHT11_TIMEOUT_ERROR;
  // Lecture des 40 bits de données du capteur.
  for (byte i = 0; i < 40; ++i) {
    // Attente d'un état LOW.
    timeout = 0;
    while(!(*in & bit))
    if (++timeout == 10000)
    return DHT11_TIMEOUT_ERROR;
    // Mise en mémoire du "temps courant".
    unsigned long t = micros();
    // Attente d'un état HIGH.
    timeout = 0;    while(*in & bit)
    if (++timeout == 10000)
    return DHT11_TIMEOUT_ERROR;
    // Si le "delta Temps" est supérieur à 40µS c'est un "1", sinon c'est un "0".
    if ((micros() - t) > 40)
    Variable_DATA[Index] |= (1 << BIT_serie); // "1"
    // Le tableau Variable_DATA[] est initialisé à "0" par défaut si le compteur de bits atteint zéro.
    if (BIT_serie-- == 0) {
      BIT_serie = 7; // On passe à l'octet suivant.
      ++Index;}}
  /* Format des données :
   * [0] = Humidité en %
   * [1] = zéro
   * [2] = Température en degrés Celsius
   * [3] = zéro
   * [4] = checksum (humidité + température)   */
  *Humidite = Variable_DATA[0];
  *Temperature = Variable_DATA[2];
  // >>> Vérifie la totalisation de contrôle. (Checksum) <<<
  if (Variable_DATA[4] != (Variable_DATA[0] + Variable_DATA[2]))
  return DHT11_CHECKSUM_ERROR;
  // Erreur de checksum.
  else    return DHT11_SUCCESS;} // Pas d'erreur.
//----------------------------------------------------------------------------------------------
void AFF_etoile() {Serial.print('*');}
  
void Attendre_une_chaine_sur_RX() {// Cette procédure attend une chaine sur la ligne USB.
  num_ERR = 0; Sentinelle_detectee = false;  // Sentinelle non reçue.
  Compteur_caracteres = 0; Chaine_Memorisee = ""; // Vider le tampon pour une nouvelle capture.
  while (!Sentinelle_detectee) {
    if (Serial.available()) {Caractere = Serial.read(); // Puiser un caractère dans le FIFO.
    // Puis concaténer ce caractère y compris la sentinelle :
    if (Compteur_caracteres < LGR_chaine) 
       {Chaine_Memorisee = Chaine_Memorisee + Caractere; Compteur_caracteres++;}
    // Si sentinelle détectée, informer la boucle de réception :
    Sentinelle_detectee = (Caractere == '*');}}
  Code_Consigne = Chaine_Memorisee.toInt();
  Repondre = true;}
  
 void BIP() {
  pinMode(BUZZER, OUTPUT); tone(BUZZER,3000); delay(100); noTone(BUZZER); pinMode(BUZZER, INPUT);}

