/* Programme Démonstrateur directement issu de la version H pour mettre au point sur le
   PUPITRE diverses fonctions d'exploitation.
   Programme à associer avec P22J_Demonstrateur_Sonde.ino
   
   ATTENTION : Pour que les textes affichés soient cohérents, à partir de ce démonstrateur il faut
               impérativement utiliser P35_Ecrire_les_textes_en_EEPROM.ino pour les inscrire en EEPROM.
   
   UTILISATION DU PROGRAMME :
 ----------------------------
 1) Établir la liaison TX/RX avec l'inverseur,
 2) Faire un RESET sur la Raquette,
 3) Cliquer sur une touche pour passer à l'exploration du clavier.
 4) Cliquer à convenance sur toutes les touches du clavier.
    Si ERR ou version du PGM un BIP est déclenché.
 5) Tourner à convenance le codeur incrémental.
 La LED rouge A0 sur la sonde : Rapide allume et Lent éteint la LED.
   
 Ce que fait ce démonstrateur :
 1) Active / Désactive la SECURITÉ. (État indiqué par une LED rouge dédiée pilotée par A2.)
 2) Il permet de sélectionner l'un des neuf programmes enregistrés en EEPROM.
 3) Il permet d'effacer le programme indexxé en EEPROM si la SECURITE est armée.
 4) Après SAV Balayage, Spectre coloré et Posture affiche le résultat.
 5) Permet l'effacement du dernier code d'un PGM enregistré. (B.S.)
 6) Exécuter le PGM indexé.
 7) Editer le programme indexé en EEPROM.
 8) Déplacer un programme en EEPROM.
 9) Le bouron central du codeur rotatir en EXPLOITATION et sous condition impose de faire un délai d'une seconde.
 
                                                             PGM            RAM dynamique 
Démonstrateur sans modification :                      16488 Octets (53%)   354 Octets (17%)
Active / Désactive la SECURITÉ :                       16506 Octets (53%)   354 Octets (17%)   +98 octets
Sélection de l'un des huit programmes en EEPROM :      16880 Octets (54%)   368 Octets (17%)  +374 octets
Effacer le programmes indexé en EEPROM :               17032 Octets (55%)   372 Octets (18%)  +152 octets
Lister le programmes indexé en EEPROM :                17562 Octets (57%)   380 Octets (18%)  +530 octets
SAV Balayage, Luxmètre et Posture Aff le résultat :    17572 Octets (57%)   380 Octets (18%)   +10 octets
Effacement du dernier code d'un PGM enregistré :       17646 Octets (57%)   384 Octets (18%)   +74 octets
Correction Effacement PGM et Exécuter le PGM indexé
et Editer le programme indexé en EEPROM :              18046 Octets (58%)   400 Octets (19%)  +400 octets
Déplacer un programme en EEPROM :                      18168 Octets (59%)   404 Octets (19%)  +122 octets
Diverses optimisations :                               18080 Octets (58%)   404 Octets (19%)   -88 octets
Modif textes / optimisations et code pour Délai 1S :   18160 Octets (59%)   404 Octets (19%)   +80 octets

//--------------------- Déclarations pour le matériel ----------------------- */
#include <EEPROM.h>
#include <Adafruit_ssd1306syp.h> // Afficher graphique OLED.
// Déclarations pour le capteur codeur rotatif KY-40 :
#define Codeur_A 2    // D2 : DT
#define Codeur_B 3    // D3 : CLK
// Définition des broches pour la ligne I2C.
#define broche_SDA 11 // D11 utilisée en sortie.
#define broche_SCL 12 // D12 utilisée en sortie.
// Initialisation des broches en sorties et affectations.
Adafruit_ssd1306syp display(broche_SDA, broche_SCL);
//--------------- Définition des broches d'Entrées Analogiques --------------
#define Mesure_U_clavier 6 // Entrée analogique A6 utilisée pour le clavier.
#define BP_central 21 // A7 pour lire le bouton central.
//------------------ Sorties sur broches "Analogiques" ----------------------
#define LED_SECURITE 16 // Sortie A2 pilote la LED ROUGE de SECURITÉ.
#define LED_MOUVEMENTS 17 // Sortie A3 pilote la LED BLEUE : MOUVEMENTS.
#define LED_Retour_par_FIN 18 // Sortie A4 pilote la LED Retour par FIN.
#define LED_Clavier 19 // Sortie A5 pilote la LED Clavier.
//-------------- Définition des broches de sorties Binaires -----------------
#define LED_Erreur 4 // LED rouge : MOUVEMENTS ou ERREUR.
#define BUZZER 6
#define C1 7  // Colonne C1 sur la sortie D7.
#define C2 8  // Colonne C1 sur la sortie D8.
#define C3 9  // Colonne C1 sur la sortie D9.
#define C4 10 // Colonne C1 sur la sortie D10.
#define LED_Loop_active 13 // D13 utilisée pour la LED clignotante "Boucle de base".
//------------------------ Constantes du programme --------------------------
#define Version_PGM "09-02-2018"    //@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
#define Num_initial_donnees 1       //@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
#define Num_initial_options 7       //@@@@@@@ (7 en version définitive.) @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
#define Num_initial_Exploitation 4  //@@@@@@@ (4 en version définitive.) @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
#define Num_initial_des_postures 9  //@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

#define LGR_chaine 20 //@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//------------------------- Variables du programme --------------------------
// Note : Les déclarations sont classées par type et par ordre alphabétique croissant.
boolean Change_le_num_PGM_cible;
boolean Gradateur; // Si true le Bouton rotatif agit sur le gradateur, si false agit sur les Items des menus.
boolean LED_Erreur_clignote;
boolean LED_Retour_par_FIN_clignote;
boolean LASER; // Allumer/éteindre.
boolean Mode_pointage_LASER;
boolean Mode_MANUEL;
boolean PHARES;
boolean Sens_positif; // Sens de rotation.
boolean Sentinelle_detectee; // Délimiteur de fin détecté.
boolean Verifier_OK; // Vérifier que l'ACR est bien "OK*".
boolean Vertical; // Le LASER est orienté verticalement. (Consigne potentiométrique.)

byte Articulation;
byte BP; // Bouton du claviar activé.
byte BIT_Option;
byte Code_Consigne;
byte Compteur_caracteres;
byte Etat_B;
byte Mode_en_cours; // 1 : EXPLOITER.   2 : Imposer des POSTURES. (Correspondent au BP du clavier.)
                    // 3 : OPTIONS de la sonde.   4 : DONNEES de la sonde.    5: MOUVOIR.
byte NB_a_incrementer_decrementer;                    
byte NB_Repetition_Mvt;
byte Num_des_donnees;
byte Num_Exploitation;
byte Num_des_options;
byte Num_des_postures;
byte Num_PGM;
byte RETOUR; // Code de retour d'une fonction à risque. (Code 0 : pas de retour nécessaire.)
byte Transition;

char Jambe;
char Caractere;

int CNA; // Mesure analogique retournée par le CNA.

unsigned long Ancienne_valeur_Temps_Ecoule_depuis_RESET;

String Chaine_Memorisee; // Tampon mémoire pour la chaine entrante.

void setup()  {
  //------------------------ Initialisations pour le matériel --------------------------
  display.initialize(); delay(300); display.clear();
  Serial.begin(115200);
  //------------------------------ Affiche écran initial -------------------------------
  display.setTextSize(1); // Taille la plus petite.
  Affiche_cadre_du_haut(11); Aff_TEXTE_EEPROM(202,8); display.print(Version_PGM);
  CADRE(); display.setTextSize(2); // Taille double.
  display.setCursor(38,24); Aff_TEXTE_EEPROM(0,5);
  display.setCursor(32,41); AFF_mot_JEKERT();
  display.update(); display.setTextSize(1); // Taille la plus petite.
  //-------------------------- Initialisations des booléens ----------------------------
  PHARES = false; LASER = false; Gradateur = false; RETOUR = 0; LED_Retour_par_FIN_clignote = false;
  Mode_pointage_LASER = false; Vertical = true; LED_Erreur_clignote = false; Mode_MANUEL = false;
  Change_le_num_PGM_cible = false;
  //------------------------- Initialisations des variables ----------------------------
  NB_Repetition_Mvt = 1; Jambe = 'A'; Articulation = 1; 
  //------------------------------ Initialisations des E/S -----------------------------
  pinMode(C1,OUTPUT); pinMode(C2,OUTPUT); pinMode(C3,OUTPUT); pinMode(C4,OUTPUT); digitalWrite(C1,LOW);
  digitalWrite(C2,LOW); digitalWrite(C3,LOW); digitalWrite(C4,LOW); pinMode(LED_Clavier,OUTPUT); 
  pinMode(LED_Erreur,OUTPUT); digitalWrite(LED_Erreur,LOW); pinMode(Codeur_A,INPUT); pinMode(Codeur_B,INPUT);
  pinMode(LED_Retour_par_FIN,OUTPUT); digitalWrite(LED_Retour_par_FIN,LOW);
  pinMode(LED_MOUVEMENTS,OUTPUT); digitalWrite(LED_MOUVEMENTS,LOW);
  pinMode(LED_SECURITE,OUTPUT); digitalWrite(LED_SECURITE,LOW);
  //------------------------------------------------------------------------------------
  Chaine_Memorisee.reserve(LGR_chaine); // Reserve LGR_chaine octets pour la chaine.
  //ATTENDRE_un_BP();  //@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
  Mode_en_cours = 1; //MENU EXPLOITER pour faire sortir la sonde de l'hibernation ...
  Num_Exploitation = Num_initial_Exploitation; Affiche_Ecran_MENU(); // Pour afficher l'item "Réveiller JEKERT". 
  Transition = LOW; }//Etat_B = LOW;} // LOW initial simule une rotation pour déclencher le premier menu.

void loop() {
  //------------- Faire clignoter la LED "boucle de base" et Mouvements ----------------
  if (millis() > Ancienne_valeur_Temps_Ecoule_depuis_RESET) {
      Ancienne_valeur_Temps_Ecoule_depuis_RESET = millis() + 100; // Réarme pour 0,1S ou 0,03S.
      digitalWrite(LED_Loop_active,!digitalRead(LED_Loop_active)); // Inverser l'état de la LED.
      if (LED_Erreur_clignote) digitalWrite(LED_Erreur,!digitalRead(LED_Erreur));
      else digitalWrite(LED_Erreur,LOW);
      if (LED_Retour_par_FIN_clignote) digitalWrite(LED_Retour_par_FIN,!digitalRead(LED_Retour_par_FIN));
      else digitalWrite(LED_Retour_par_FIN,LOW);}
  //------------------------------- Traiter le clavier ---------------------------------
  Tester_le_clavier();
  if (BP > 0) { // Si BP > 0 c'est qu'un B.P. a été activé.
      Effacer_ACR();
      //================================================================================
      if (Mode_pointage_LASER) {
          if ((BP == 9) || (BP == 11)) Envoyer_Consigne(64); // Balayage vertical.
          if ((BP == 6) || (BP == 14)) Envoyer_Consigne(65); // Balayage horizontal.
          if (BP == 8) {Envoyer_Consigne(63); Purge_les_booleens();} // Touche FIN pour sortir.
          if (BP == 10) Envoyer_Consigne(9);} // Simuler un tir au LASER.
      // NOTE : L'exploration de S15 n'est pas effectuée dans cette séquence car réalisée dans
      //        la séquence "else de balayage des touches de MENU". Il y aurait contradiction ici.
      //================================================================================
      if (Mode_MANUEL) {
          switch (BP) {
            case  5 : {Envoyer_Consigne(79); Articulation = 1; break;}
            case  8 : {Traite_la_touche_NON(); Affiche_Ecran_MENU(); display.setCursor(1,37);
                       AFF_TXT_EEPROM_et_RETOUR(691,17); display.update(); break;}
            case  9 : {Envoyer_Consigne(78); Jambe = 'D'; break;}
            case 11 : {Envoyer_Consigne(77); Jambe = 'C'; break;}
            case 12 : {Envoyer_Consigne(80); Articulation = 2; break;}
            case 13 : {Envoyer_Consigne(76); Jambe = 'B'; break;}
            case 15 : {Envoyer_Consigne(81); Articulation = 3; break;}
            case 16 : {Envoyer_Consigne(75); Jambe = 'A';}}
            if (Gradateur) Affiche_moteur_actif();}
      //================================================================================
      else {//---------------------- Touches de changement de MENU. ----------------------------------------
            if (BP < 6) {Mode_en_cours = BP; Sens_positif = true; // Pour ouvrir sur la bonne page écran.
                         Num_des_donnees = Num_initial_donnees; Num_des_postures = Num_initial_des_postures;
                         Num_des_options = Num_initial_options; Num_Exploitation = Num_initial_Exploitation; 
                         Affiche_Ecran_MENU(); Traite_Item_Menus();} // Boutons à validité permanente.
            //----------------------------------------------------------------------------------------------
            if (BP == 7) Traite_la_touche_OUI();
            if (BP == 8) Traite_la_touche_NON();
            Teste_Touche_PHARES (); // Traite la touche 13.
            if (!Mode_MANUEL) {Teste_Touche_LASER();} // Teste la touche 15 et allume ou éteind le LASER.
            //------------------------------------- Traite les MOUVEMENTS ----------------------------------
            if (Mode_en_cours == 5) {// Mode MOUVOIR
                for (byte I = NB_Repetition_Mvt; I>0; I--)
                    switch (BP) { // Traite les touches de déplacement.
                      case 6  : {Envoyer_Consigne(7); break;} // Translation à droite.
                      case 9  : {Envoyer_Consigne(3); break;} // Avancer.
                      case 10 : {Envoyer_Consigne(14); break;} // Libérer les efforts.
                      case 11 : {Envoyer_Consigne(4); break;} // Reculer.
                      case 12 : {Envoyer_Consigne(5); break;} // Tourner à droite.
                      case 14 : {Envoyer_Consigne(8); break;} // Translater à gauche.
                      case 16 :  Envoyer_Consigne(6);}}}}      // Tourner à gauche.
  //----------------------------- Traite le B.P. central -------------------------------
  if (analogRead(BP_central) < 512) { // Si le B.P. central est activé :
      digitalWrite(LED_Clavier,HIGH);
      while (analogRead(BP_central) < 512); // Attendre le relâché.
      digitalWrite(LED_Clavier,LOW);
      //---------------------------------------------------
      if (Mode_en_cours == 1) // Mode EXPLOITATION.
          switch (Num_Exploitation) {
            case 11  : {Envoyer_Consigne(66); break;} // Inverser l'amplitude du débattements si Exploitation LASER.
            case 15  : {Envoyer_Consigne(66); break;} // Inverser l'amplitude du débattements si pilotage Manuel.
             default : Envoyer_Consigne(90);} // Générer une temporisation d'une seconde.
      if (Mode_en_cours == 2) Envoyer_Consigne(55); // Active la sonde si mode POSTURES.
      if (Mode_en_cours == 3) digitalWrite(LED_SECURITE,HIGH);} // Arme la sécurité si mode OPTIONS.
  //------------------------------ Traite la rotation ----------------------------------
  Etat_B = digitalRead(Codeur_B);  
  if ((Transition == LOW) && (Etat_B == HIGH)) {// Rotation détectée.
       if (digitalRead(Codeur_A) == LOW) Sens_positif = true; else Sens_positif = false;
       // ATTENTION : l'instruction display.update() fausse le test ci-dessus. Il est impératif de
       // placer Effacer_ACR() après avoir positionné Sens_positif.
       Efface_le_code(); Effacer_ACR(); // Efface Code TX et ACR.
       Afficher_Rot_ou_CLV(); // Pour effacer un éventuel message d'erreur.
       //-------------------------------------------------------------------------------
       if (Gradateur) Traite_Gradateur(); // Si NON : Traite les menus ci-dessous.
       else {if (Mode_en_cours == 1) Effacer_zone(6,55,16); // Effacer "Retour par FIN" si mode EXPLOITER.
             if (Mode_en_cours == 5) {// Menu MOUVOIR.
                 NB_a_incrementer_decrementer = NB_Repetition_Mvt; Incrementer_Decrementer(9);
                 NB_Repetition_Mvt = NB_a_incrementer_decrementer;
                 Afficher_le_NB_fois();}
             Traite_Item_Menus(); display.update();}}
       //-------------------------------------------------------------------------------
  Transition = Etat_B;}

//-------------------------------------- Traitement du CLAVIER -------------------------------------
void Traite_la_touche_OUI() {
  if (Mode_en_cours == 1) { // Menu EXPLOITER.
      switch (Num_Exploitation) {
        case  1 : {Envoyer_Consigne(69); Affiche_le_balayage_ultrasons(); break;} // Enregistrer balayage ultrasons.     
        case  2 : {Affiche_le_balayage_ultrasons(); break;} //Afficher la télémétrie.
        case  3 : {Envoyer_Consigne(18); AFF_POSTURE(83); break;} // Enregistrer en EEPROM la posture actuelle.
        case  4 : {Envoyer_Consigne(61); digitalWrite(LED_SECURITE,LOW);
                   LASER = false; break;}  // Passer la Sonde eu Hibernation. (QUITTER.)
        case  5 : {Envoyer_Consigne(55); break;} // Réveiller la sonde.
        case  6 : {Envoyer_Consigne(58); break;} // Test de communication MPU6050.
        case  7 : {Envoyer_Consigne(59); Affiche_Calage_Gyroscope(1,37); break;} // Calage du gyroscope.
        case  8 : {Envoyer_Consigne(54); break;} // LASER et Phares forcés à 127.
        // Signale à la sonde que le gradateur agit sur les phares :
        case  9 : {Envoyer_Consigne(43); Gradateur = true; AFFICHE_puissance_eclairages();
                   LED_Retour_par_FIN_clignote = true; break;}
        // Signale à la sonde que le gradateur agit sur le LASER :
        case 10 : {Envoyer_Consigne(42); Gradateur = true; AFFICHE_puissance_eclairages();
                   LED_Retour_par_FIN_clignote = true; break;}
        case 11 : {Envoyer_Consigne(62); Mode_pointage_LASER = true; LED_Retour_par_FIN_clignote = true;
                   LED_Erreur_clignote = true; Gradateur = true; LASER = false; break;}
        case 12 : {Envoyer_Consigne(67); AFF_MENU_Spectre_colorimetrique(); break;} // Enregistrer colorimétrique.
        case 13 : {AFF_MENU_Spectre_colorimetrique(); break;} // Visualiser un spectre colorimétrique.
        case 14 : {LED_Retour_par_FIN_clignote = true; LED_Erreur_clignote = true; // Mvt Torsion. 
                   Envoyer_Consigne(38); RETOUR = 39; Gradateur = true; break;} // Pour BR envoie (40) & (41).   
        case 15 : {LED_Retour_par_FIN_clignote = true; LED_Erreur_clignote = true; // Pilotage MANUEL. 
                   Envoyer_Consigne(73); RETOUR = 74; Gradateur = true; Mode_MANUEL = true;
                   Affiche_moteur_actif(); Envoyer_Consigne(79); Envoyer_Consigne(75); break;}
        case 16 : {AFF_POSTURE(82); break;} // Aff. posture Actuelle.
        case 17 : {AFF_POSTURE(83); break;} // Aff. posture SAV en EEPROM.
        case 18 : {Envoyer_Consigne(9); break;} // Tir LASER.
        case 19 : {Effacer_un_PGM(); break;} // Efface le PGM indexé en EEPROM.
        case 20 : {Lister_un_PGM(); break;} // Liste le PGM indexé en EEPROM.
        case 21 : {Effacer_le_dernier_code_appris(); break;} // Effacer le dernier code enregistré.
        case 22 : {Envoyer_Consigne(20); break;} // Editer le PGM indexé en EEPROM.        
        case 23 : {Executer_un_PGM_appris(); break;} // Déclencher le programme indexé.
        case 24 : {Envoyer_Consigne(12); break;} // Le PGM indexé devient la SOURCE d'un transfert.
        case 25 : {Transferer_la_SOURCE(); break;}}} // La SOURCE passe dans le PGM actuel et est effacée.    
  if (Mode_en_cours == 2)  { // Menu des POSTURES. 
      switch (Num_des_postures) {
        case 1 : {Envoyer_Consigne(1); break;}  // Posture VEILLE. (Sonde posée au sol.)
        case 2 : {Envoyer_Consigne(16); RETOUR = 17; LED_Retour_par_FIN_clignote = true; break;} // STBRaisonnable.
        case 3 : {Envoyer_Consigne(10); LED_Retour_par_FIN_clignote = true; RETOUR = 11; break;} // Hauteur MAXI.
        case 4 : {Envoyer_Consigne(15); break;} // Moteurs au Neutre opérationnel.
        case 5 : {Envoyer_Consigne(19); break;} // Imposer la posture apprise en EEPROM.
        case 6 : {Envoyer_Consigne(23); break;} // Configuration décollage.
        case 7 : {Envoyer_Consigne(22); break;} // Atterrissage.
        case 8 : {Envoyer_Consigne(21); break;} // Déposer la sonde.
        case 9 : Envoyer_Consigne(2);}}         // Stable Transversal.    
  if (Mode_en_cours == 3) { // Menu des OPTIONS.
    switch (Num_des_options) { // Change l'étét du booléen système.
      case 1 : {Envoyer_Consigne(25); break;} // Disjoncter les énergies.
      case 2 : {Envoyer_Consigne(32); break;} // Moteurs ON.
      case 3 : {Envoyer_Consigne(33); break;} // Mouvements rapides.
      case 4 : {Envoyer_Consigne(35); break;} // Stabilisation Gyroscopique.
      case 5 : {if (digitalRead(LED_SECURITE)) { // Si sécurité activée :
                    if (Apprentissage_en_cour()) // Valider le mode apprentissage si non actif.
                       {Chaine_Memorisee = "E11"; BIP_et_XY();  display.print("E11");}
                    else Envoyer_Consigne(71);}
                else ERR_NON_Securite(); break;}
      case 6 : {Changer_de_PGM_EEPROM(); break;}
      case 7 : {digitalWrite(LED_SECURITE,HIGH); break;}} // Arme la sécurité.
      Affiche_etat_option_en_cours();}
  display.update();}

void Traite_la_touche_NON() { // Touche NON et touche FIN.
  Purge_les_booleens(); // Pour que le BR agisse sur les MENUs.
  //------------------------------------------------------------------------------- 
  if (RETOUR > 0) {Envoyer_Consigne(RETOUR); RETOUR = 0;} // Configuration de retour des fonctions à risque.
  //------------------------------------------------------------------------------- 
  if (Mode_en_cours == 1) { // Menu EXPLOITATION. : Signale la fin du gradateur par le code 53.
      Envoyer_Consigne(53);
      if (!Mode_MANUEL) display.update(); // Efface la visualisation des valeurs ligne du bas.
      // Codes 16 et 17 : Affichage de la posture actuelle ou de la posture SAV en EEPROM.
      if ((Num_Exploitation == 16) || (Num_Exploitation == 17)) Regenere_le_menu_posture();}
  //-------------------------------------------------------------------------------         
  if (Mode_en_cours == 3) { // Menu des OPTIONS.
      switch (Num_des_options) { // Change l'étét du booléen système.
        case 1 : {Envoyer_Consigne(26); break;} // Disjoncteur énergie sur OFF.
        case 2 : {Envoyer_Consigne(31); break;} // Moteurs ON.
        case 3 : {Envoyer_Consigne(34); break;} // Mouvements lents.
        case 4 : {Envoyer_Consigne(36); break;} // Plus de stabilisation Gyroscopique.
        case 5 : {Envoyer_Consigne(72); break;} // Fin du Mode apprentissage.
                 // Ci-dessous : Fin du changement de PGM indexé en EEPROM.
        case 6 : {Gradateur = false; Change_le_num_PGM_cible = false; LED_Retour_par_FIN_clignote = false;}
        case 7 : {digitalWrite(LED_SECURITE,LOW); break;}} // Coupe la sécurité.
        Affiche_etat_option_en_cours();}
  //-------------------------------------------------------------------------------
  digitalWrite(LED_Retour_par_FIN,LOW);}

//-------------------------- Gestion de l'apprentissage et de son l'utilisation --------------------

void Realiser_un_PGM() {
  Effacer_ACR(); // Pour contrôler le "OK" de la fin d'exécution du programme.
  //--------------------- Déclencher la réalisation du PGM ------------------------
  if (digitalRead(LED_SECURITE)) {
      Demande_Information(87); // Déclenche le programme sans vérifier un retour "OK".
      if (Chaine_Memorisee != "E18") {
      //--------------- Barrer les codes dans la grille de listage ----------------
      byte C=1; byte L=26;
      while ((Chaine_Memorisee != "OK") && (Chaine_Memorisee != "E20")) {
             display.drawLine(C, L, C+16, L-11, WHITE); display.update(); // barre la case et raffraichi l'écran.
             L = L+12; if (L > 62) {L = 26; C = C+16;} // Change de case.
             Attendre_une_chaine_sur_le_Maitre();} // Attente du code suivant.
             } digitalWrite(LED_SECURITE,LOW); AFF_ACR();}
  else ERR_NON_Securite();}
  
void ERR_NON_Securite() {BIP_et_XY(); display.print("E12");}

void BIP_et_XY() {BIP(); display.setCursor(4,4);}

boolean Apprentissage_en_cour() {
  Demande_Information(44); // Retourne l'état des booléens système.
  return Chaine_Memorisee.charAt(7) == '1';}

void Executer_un_PGM_appris() {
  if (!Apprentissage_en_cour()) {Lister_un_PGM_si_SECURITE(); Realiser_un_PGM();}
  else {BIP_et_XY(); display.print("E19");}}

void Transferer_la_SOURCE() {
  if (!Apprentissage_en_cour()) {
      if (digitalRead(LED_SECURITE)) {Envoyer_Consigne(13); digitalWrite(LED_SECURITE,LOW);}
      else ERR_NON_Securite();}
  else {BIP_et_XY(); display.print("E22");}}
  
void Effacer_un_PGM() {
  if (!Apprentissage_en_cour()) {
      if (digitalRead(LED_SECURITE)) {Envoyer_Consigne(88); digitalWrite(LED_SECURITE,LOW);}
      else ERR_NON_Securite();}
  else {BIP_et_XY(); display.print("E21");}}
  
void Effacer_le_dernier_code_appris() {
  if (Apprentissage_en_cour()) Envoyer_Consigne(89);
  else {digitalWrite(LED_SECURITE,LOW); BIP_et_XY(); display.print("E16");}}
  
void Lister_un_PGM_si_SECURITE() {if (digitalRead(LED_SECURITE)) Lister_un_PGM();}
  
void Lister_un_PGM() {
  byte C=3; byte L=18;
  Prepare_afficher_item(); // Efface la ligne centrale de texte.
  //----------------------- Affiche la grille de listage --------------------------
  CADRE(); for (byte C=16; C<113; C=C+16) display.drawLine(C, 17, C, 62, WHITE);
  for (byte L=15; L<63; L=L+12) display.drawLine(1, L, 127, L, WHITE);
  //------------------------- Affiche le n° du PGM ciblé --------------------------
  Demande_Information(84); display.drawRect(114, 41, 13, 9, WHITE);
  display.drawRect(113, 40, 14, 11, WHITE);
  display.setCursor(115,42); display.setTextColor(BLACK, WHITE); display.print('P');
  display.print(Chaine_Memorisee); display.setTextColor(WHITE, BLACK); 
  //---------------------- Réceptionne et liste les codes -------------------------
  Serial.print("86*");
  while (!Fin_de_listage()) {
         Attendre_une_chaine_sur_le_Maitre(); display.setCursor(C,L);
         if (!Fin_de_listage()) 
             {if (Chaine_Memorisee.length() < 2) display.print('0');
                  display.print(Chaine_Memorisee); display.update();}
         L = L+12; if (L > 62) {L = 18; C = C+16;}}
   if (Chaine_Memorisee == "E15") {BIP_et_XY(); display.print("E15");}}
  
boolean Fin_de_listage() {return ((Chaine_Memorisee == "OK") || (Chaine_Memorisee == "E15"));} 

void Affiche_Num_PGM_cible() {
  Demande_Information(84); display.print(Chaine_Memorisee); ESP_Point_Int();} 
  
void Changer_de_PGM_EEPROM() {
  AFF_num_PGM_actif(); Gradateur = true; Change_le_num_PGM_cible = true; LED_Retour_par_FIN_clignote = true;}
  
//--------------------------------------------------------------------------------------------------
void Affiche_moteur_actif() {
  Effacer_zone(55,23,1); display.setCursor(7,23); Aff_TEXTE_EEPROM(717,8); display.print(Jambe);
  Prepare_afficher_item(); // Effacer_zone(1, 37, 21); display.setCursor(1,37)
  display.setCursor(1,37); Aff_TEXTE_EEPROM(708,9);
  switch ( Articulation) { 
    case 1 : {Aff_TEXTE_EEPROM(725,6); break;}
    case 2 : {Aff_TEXTE_EEPROM(731,5); break;}
    case 3 : Aff_TEXTE_EEPROM(736,6);}
  display.update();}
              
void Purge_les_booleens() {
  LED_Retour_par_FIN_clignote = false; LED_Erreur_clignote = false; Mode_pointage_LASER = false;
  Gradateur = false; Mode_MANUEL = false;}
    
void Teste_Touche_PHARES (){
  if (BP == 13) {PHARES = !PHARES; if (PHARES) Envoyer_Consigne(27); else Envoyer_Consigne(28);}}  
 
//-------------------------------- Récupération des données de POSTURE -----------------------------
void AFF_POSTURE(byte Code_consigne) {
  Prepare_afficher_item(); Demande_Information(Code_consigne);
  Recupere_et_affiche_la_posture(8,20); Recupere_et_affiche_la_posture(8,37); 
  Recupere_et_affiche_la_posture(8,55); Recupere_et_affiche_la_posture(41,20);
  Recupere_et_affiche_la_posture(41,37); Recupere_et_affiche_la_posture(41,55);
  Recupere_et_affiche_la_posture(72,20); Recupere_et_affiche_la_posture(72,37);
  Recupere_et_affiche_la_posture(72,55); Recupere_et_affiche_la_posture(104,20);
  Recupere_et_affiche_la_posture(104,37); Efface_le_code();
  display.setCursor(104,55); display.print(Chaine_Memorisee);
  LED_Retour_par_FIN_clignote = true;}
  
void Recupere_et_affiche_la_posture (byte X, byte Y) {
  display.setCursor(X,Y); display.print(Chaine_Memorisee);
  Attendre_une_chaine_sur_le_Maitre();}
  
void Regenere_le_menu_posture() {
  Prepare_afficher_item(); Effacer_zone(8,20,20); Effacer_zone(8,55,20); Efface_le_code();
  if (Num_Exploitation == 16) Aff_TEXTE_EEPROM(742,18); else Affiche_item_posture_EEPROM();
  display.update();}
  
//-------------------------- Gestion de l'affichage du balayage télémétrique ----------------------- 
void Affiche_le_balayage_ultrasons() {
  Prepare_afficher_item(); CADRE(); display.update(); Demande_Information(70);
  RETOUR = 23; // Position verticale du tracé. (On réutilise cette variable disponible durant cette procédure.)
  for (byte I = 0; I<31; I++) {// Lecture d'une donnée et tracé de sa ligne représentative.
       Affiche_une_ligne_representative(); Attendre_une_chaine_sur_le_Maitre();}
  Affiche_une_ligne_representative(); display.update(); }
 
 void Affiche_une_ligne_representative() {
   CNA = Chaine_Memorisee.toInt(); CNA = CNA * 2; // Pour une portée de 2m sur la largeur d'affichage.
   for (CNA; CNA > 0; CNA--) display.drawPixel(CNA+5, RETOUR, WHITE); RETOUR++;}

//-------------------------------- Gestion du spectre colorimétrique -------------------------------  
void AFF_MENU_Spectre_colorimetrique() {
  Prepare_afficher_item(); Demande_Information(68);
  Recupere_et_affiche_couleur(5,18,635,17); // Lumiere totale.
  Recupere_et_affiche_couleur(5,30,309,6); // Bleu.
  Recupere_et_affiche_couleur(5,42,315,6); // Mauve.
  Recupere_et_affiche_couleur(5,54,321,6); // Rouge.
  Recupere_et_affiche_couleur(71,30,327,6); // Orange.
  Recupere_et_affiche_couleur(71,42,622,6); // Jaune. 
  Efface_le_code(); display.setCursor(71,54); Aff_TEXTE_EEPROM(628,6); display.print(Chaine_Memorisee);} // Vert.
  
void Recupere_et_affiche_couleur(byte X, byte Y, int PTR, byte Lgr) {
  display.setCursor(X,Y); Aff_TEXTE_EEPROM(PTR,Lgr); display.print(Chaine_Memorisee);
  Attendre_une_chaine_sur_le_Maitre();}

//---------------------------------------- Gestion des MENUs --------------------------------------- 
void MENU_POSTURES(){
  NB_a_incrementer_decrementer = Num_des_postures; Incrementer_Decrementer(9);
  Num_des_postures = NB_a_incrementer_decrementer;
  Affiche_Ecran_POSTURES(); display.update();}
  
void AFF_mot_JEKERT() {Aff_TEXTE_EEPROM(552,6);} 
  
void Affiche_Ecran_POSTURES(){
  display.setCursor(23,21); Aff_TEXTE_EEPROM(30,13); Effacer_zone(6, 55, 17);
  Prepare_afficher_item();
  switch (Num_des_postures) {
    case 1 : {AFF_mot_JEKERT(); display.print(' '); display.print(char(25));  // La sonde se pose sur le sol.
              Aff_TEXTE_EEPROM(285,7); break;}
    case 2 : {AFF_TXT_EEPROM_et_RETOUR(452,18); break;} // Stable Raisonnable.
    case 3 : {AFF_TXT_EEPROM_et_RETOUR(470,16); break;} // Hauteur Maximale.
    case 4 : {Aff_TEXTE_EEPROM(486,20); break;} // Moteurs au Neutre opérationnel.
    case 5 : {Aff_TEXTE_EEPROM(506,17); break;} // Restituer la posture apprise en EEPROM.
    case 6 : {BIP(); Aff_TEXTE_EEPROM(535,9); Aff_TEXTE_EEPROM(602,12); break;}  // Décollage.
    case 7 : {Aff_TEXTE_EEPROM(523,12); break;} // Atterrissage.
    case 8 : {Aff_TEXTE_EEPROM(544,14); break;} // Déposer la Sonde.
    case 9 : {Aff_TEXTE_EEPROM(558,18);}}       // Stable Transversal.
    if (Num_des_postures != 6) display.print('.');}
     
void MENU_EXPLOITER(){
  NB_a_incrementer_decrementer = Num_Exploitation; Incrementer_Decrementer(25);
  Num_Exploitation = NB_a_incrementer_decrementer;
  Affiche_Ecran_EXPLOITER(); display.update();}
 
void Affiche_Ecran_EXPLOITER(){
  Prepare_afficher_item();
  switch (Num_Exploitation) {
    case  1 : {Aff_TEXTE_EEPROM(333,3); Aff_TEXTE_EEPROM(661,14); break;} // Enregistrer un balayage télémètre.   
    case  2 : {Aff_TEXTE_EEPROM(354,3); Aff_TEXTE_EEPROM(661,14); break;} // Afficher le balayage télémétrique.
    case  3 : {Aff_TEXTE_EEPROM(251,19); break;} // Sauvegarder la posture actuelle.
    case  4 : {Aff_TEXTE_EEPROM(270,7); ESP_Point_Int(); break;} // QUITTER.
    case  5 : {Aff_TEXTE_EEPROM(393,10); AFF_mot_JEKERT(); break;} // Réveiller la sonde.    
    case  6 : {Aff_TEXTE_EEPROM(389,4); Aff_TEXTE_EEPROM(167,11); break;} // Test communication avec le MPU6050.
    case  7 : {Affiche_Calage_Gyroscope(1,37); break;}  // Calage du gyroscope.    
    case  8 : {Aff_TEXTE_EEPROM(426,20); break;} // Phare et LASER = 127.    
    case  9 : {Aff_TEXTE_EEPROM(403,17); break;} // Gradateur Phare.
    case 10 : {Aff_TEXTE_EEPROM(403,10); Aff_TEXTE_EEPROM(420,6); break;} // Gradateur LASER.
    case 11 : {Aff_TEXTE_EEPROM(378,11); Aff_TEXTE_EEPROM(372,6); // Utiliser le LASER.
               ESP_Point_Int(); AFFICHE_RETOUR(); break;}
    case 12 : {Aff_TEXTE_EEPROM(333,21);break;}  // Enregistrer un spectre colorimétrique.
    case 13 : {Aff_TEXTE_EEPROM(354,3); Aff_TEXTE_EEPROM(336,18); break;} // Afficher le spectre couleur.
    case 14 : {AFF_TXT_EEPROM_et_RETOUR(675,16); break;} // TORSION active.
    case 15 : {AFF_TXT_EEPROM_et_RETOUR(691,17); break;} // Pilotage en MANUEL.
    case 16 : {Aff_TEXTE_EEPROM(742,18); break;} // Affiche "Posture actuelle ?".
    case 17 : {Affiche_item_posture_EEPROM(); break;} // Affiche "Posture en EEPROM ?".
    case 18 : {Aff_TEXTE_EEPROM(63,4); Aff_TEXTE_EEPROM(373,5); ESP_Point_Int(); break;} // Affiche "TIR LASER ?".
    case 19 : {Aff_TEXTE_EEPROM(795,15); Affiche_Num_PGM_cible(); break;} // Effacer le PGM cible.
    case 20 : {Aff_TEXTE_EEPROM(809,7); Aff_TEXTE_EEPROM(802,8); Affiche_Num_PGM_cible(); break;} // Lister Cible.
    case 21 : {Aff_TEXTE_EEPROM(816,21); break;} // Effacer le dernier code d'un PGM enregistré.
    case 22 : {Aff_TEXTE_EEPROM(844,6); Aff_TEXTE_EEPROM(802,8); Affiche_Num_PGM_cible(); break;} // Editer le PGM.
    case 23 : {Aff_TEXTE_EEPROM(837,7); Aff_TEXTE_EEPROM(802,8); Affiche_Num_PGM_cible(); break;} // Activer le PGM.
               // Ci-dessous : Le PGM actuel devient la SOURCE d'un transfert.    
    case 24 : {Aff_TEXTE_EEPROM(806,4); Demande_Information(84); display.print(Chaine_Memorisee);
               Aff_TEXTE_EEPROM(714,3); AFF_le_mot_SOURCE(); ESP_Point_Int(); break; }
    case 25 : {Aff_TEXTE_EEPROM(858,5); AFF_le_mot_SOURCE(); Aff_TEXTE_EEPROM(5,6);
               Affiche_Num_PGM_cible();break;}}} // Transférer la SOURCE dans le PGM actuellement indexé.
               
void AFF_le_mot_SOURCE() {Aff_TEXTE_EEPROM(850,6);}

void Affiche_item_posture_EEPROM() {Aff_TEXTE_EEPROM(742,8); Aff_TEXTE_EEPROM(514,9); ESP_Point_Int();}

void AFF_TXT_EEPROM_et_RETOUR(int PTR, byte LGR) {Aff_TEXTE_EEPROM(PTR,LGR); AFFICHE_RETOUR();}
  
void MENU_des_DONNEES(){
  NB_a_incrementer_decrementer = Num_des_donnees; Incrementer_Decrementer(5);
  Num_des_donnees = NB_a_incrementer_decrementer;
  Effacer_zone(7, 19, 19); Effacer_zone(5, 32, 20); Effacer_zone(5, 42, 20); Effacer_zone(5, 52, 20);
  CADRE(); display.drawLine(1, 28, 126, 28, WHITE);
  Affiche_Ecran_des_DONNEES(); display.update();}
  
void AFF_num_PGM_actif() {
  Demande_Information(84); // demande le n° de PGM actif.
  Num_PGM = Chaine_Memorisee.toInt();
  Chaine_Memorisee = "P:" + Chaine_Memorisee; AFF_ACR();}
 
void Affiche_Ecran_des_DONNEES(){
  AFF_num_PGM_actif();
  switch (Num_des_donnees) {
              // Ci dessous : Page Phares et LASER.
    case 1 : {display.setCursor(15,19); Aff_TEXTE_EEPROM(760,16);
              display.setCursor(19,32); Aff_TEXTE_EEPROM(363,15);
              Demande_Information(44); // Retourne l'état des booléens système.
              display.setCursor(25,42); OUI_NON(5); display.setCursor(85,42); OUI_NON(6);
              Demande_Information(46); display.setCursor(25,52); display.print(Chaine_Memorisee);
              Demande_Information(47); display.setCursor(85,52); display.print(Chaine_Memorisee); break;}
              // Version du PGM de la Sonde, État motorisation et état Hibernation.
    case 2 : {Demande_Information(24); // Version du PGM de la Sonde.
              display.setCursor(10,19); Aff_TEXTE_EEPROM(202,8); display.print(Chaine_Memorisee);             
              Demande_Information(50); display.setCursor(5,32); AFF_mot_JEKERT();Aff_TEXTE_EEPROM(651,11);
              display.print(Chaine_Memorisee); // État hibernation.              
              Demande_Information(44); // Retourne l'état des booléens système.
              display.setCursor(5,42); Aff_TEXTE_EEPROM(220,17); OUI_NON(1);        
              display.setCursor(23,52); Aff_TEXTE_EEPROM(237,14); OUI_NON(2); break;}
              // Ci dessous : Page  Télémètre, Météo et Eclairages O/N.
    case 3 : {Curseur_7_19(); Demande_Information(51); display.print(Chaine_Memorisee);
              display.setCursor(7,32); Demande_Information(48); // Retourne la valeur de la Température.              
              display.print(Chaine_Memorisee); Affiche_Degre(); display.print('C');
              Demande_Information(49); // Retourne la valeur de l'hygrométrie relative.
              Curseur_7_42(); display.print(Chaine_Memorisee);
              Demande_Information(44); // Retourne l'état des booléens système.
              Curseur_7_52(); Affiche_un_Soleil(); Aff_TEXTE_EEPROM(209,11); OUI_NON(0); break;}               
               // Ci-dessous : Page Distance, Stabilisation / Bouclier / Apprentissage.
    case 4 : {Demande_Information(45);
              Curseur_7_19(); display.print(Chaine_Memorisee); Demande_Information(44);
              display.setCursor(7,32); Aff_TEXTE_EEPROM(162,16); OUI_NON(3);                                          
              Curseur_7_42(); Aff_TEXTE_EEPROM(277,16); OUI_NON(4);
              Curseur_7_52(); Aff_TEXTE_EEPROM(293,16); OUI_NON(7); break;}
              // Ci dessous : Les données de navigation, CAP, Tangage, Roulis, Ecart de route.
    case 5 : {Curseur_7_19(); Demande_Information(52); // Affiche le Cap actuel.
              display.print(Chaine_Memorisee);Affiche_Degre(); Demande_Information(56);      
              display.setCursor(5,32); display.print(Chaine_Memorisee); // Affiche Tangage et Roulis.
              Affiche_Calage_Gyroscope(7,42); Curseur_7_52(); Demande_Information(60);
              display.print(Chaine_Memorisee); Affiche_Degre(); break;}}}
              
void Curseur_7_19() {display.setCursor(7,19);}
void Curseur_7_42() {display.setCursor(7,42);}
void Curseur_7_52() {display.setCursor(7,52);}

void MENU_des_OPTIONS() {
  Effacer_zone(6,55,17); // Efface le texte "RETOUR" affiché par l'option n°6.
  Effacer_zone(50,48,5); // Efface le texte (OUI) ou (NON) si option Changer le PGM indexé en EEPROM.
  NB_a_incrementer_decrementer = Num_des_options; Incrementer_Decrementer(7);
  Num_des_options = NB_a_incrementer_decrementer; Effacer_zone(5, 30, 20); display.setCursor(5,30);
  switch (Num_des_options) {
    case 1 : {Aff_TEXTE_EEPROM(115,14); display.setCursor(72,30);
              Affiche_un_Soleil(); break;}  // Mode "couper les éclairages".
    case 2 : {Aff_TEXTE_EEPROM(129,13); break;}  // Moteurs OFF.
    case 3 : {Aff_TEXTE_EEPROM(142,20); break;}  // Mouvements rapides.
    case 4 : {Aff_TEXTE_EEPROM(162,17); break;}  // Stabilisation Gyroscopique.
    case 5 : {Aff_TEXTE_EEPROM(182,20); break;}  // Mode apprentissage.
    case 6 : {AFF_TXT_EEPROM_et_RETOUR(43,19); break;} // Changer le PGM indexé en EEPROM.
    case 7 : {Aff_TEXTE_EEPROM(776,19);  break;}} // Armer / Couper la Sécurité.
  BIT_Option = Num_des_options; BIT_Option--; if (BIT_Option == 4) BIT_Option = 7;
  Affiche_etat_option_en_cours();}
   
void Affiche_Ecran_MENU() {
  Gradateur = false; digitalWrite(LED_Retour_par_FIN,LOW); digitalWrite(LED_MOUVEMENTS,LOW);
  display.clear(); CADRE_UP(); display.drawLine(100, 0, 100, 15, WHITE);
  display.drawLine(24, 0, 24, 15, WHITE); display.setCursor(34,4);
  switch (Mode_en_cours) {
    case 1 : {Aff_TEXTE_EEPROM(82,9); break;} // Exploiter.
    case 2 : {Aff_TEXTE_EEPROM(21,9); break;} // Postures.
    case 3 : {Aff_TEXTE_EEPROM(66,8); break;} // Options.
    case 4 : {Aff_TEXTE_EEPROM(74,8); break;} // Données.
    case 5 :  Aff_TEXTE_EEPROM(11,10); digitalWrite(LED_MOUVEMENTS,HIGH); Afficher_le_NB_fois();}
  Afficher_Rot_ou_CLV();} // Terminé par display.update().
     
void MENU_MOUVOIR(){
   display.setCursor(30,20); Aff_TEXTE_EEPROM(576,11); Prepare_afficher_item();
   Aff_TEXTE_EEPROM(97,18); display.update();}
  
//---------------------------------- Diverses procédures de service --------------------------------    
void Prepare_afficher_item() {Effacer_zone(1, 37, 21); display.setCursor(1,37);}
    
void ESP_Point_Int() {Aff_TEXTE_EEPROM(200,2);}

void Affiche_Degre() {display.print(char(247));}

void Affiche_un_Soleil() {display.print(char(15));}

void OUI_NON(byte Position) {
  if (Chaine_Memorisee.charAt(Position) == '1') Aff_TEXTE_EEPROM(357,3); else Aff_TEXTE_EEPROM(360,3);}

void Affiche_etat_option_en_cours() {
  if ((Num_des_options != 6) && (Num_des_options != 7))
      {Effacer_zone(56, 48, 4); Demande_Information(44); // Etat des options.
      display.setCursor(50,48); display.print('('); OUI_NON(BIT_Option); display.print(')');}
  display.update();}
  
void AFFICHE_RETOUR() {
  display.setCursor(6,55); display.print(char(16)); /* Flèche vers la droite. */ Aff_TEXTE_EEPROM(587,15);
  if (Mode_en_cours != 2) display.print('.'); BIP();} 
     
void Traite_Gradateur() {
  if (Change_le_num_PGM_cible)
     {if (Sens_positif) {Num_PGM++; if (Num_PGM == 10) Num_PGM = 1;}
      else {if (Num_PGM > 1) Num_PGM--; else Num_PGM = 9;}
      Serial.print("85P"); Serial.print(Num_PGM); Serial.print('*'); Attendre_une_chaine_sur_le_Maitre();
      AFF_num_PGM_actif();}
  else {if (Sens_positif) Envoyer_Consigne(40); else Envoyer_Consigne(41);
        if (((Num_Exploitation == 9) || Num_Exploitation == (10)) && (Mode_en_cours == 1))
              AFFICHE_puissance_eclairages();}}
      
void AFFICHE_puissance_eclairages() {
  if (Num_Exploitation == 9) Demande_Information(46); else Demande_Information(47); // Demande la puissance.
  Effacer_zone(50, 52, 3); display.setCursor(50,52); display.print(Chaine_Memorisee); display.update();}

void Affiche_Calage_Gyroscope(byte X, byte Y) {
  Effacer_zone(X+90,Y-1,4); Demande_Information(57); display.setCursor(X,Y);
  display.print(Chaine_Memorisee); Affiche_Degre();}

void Afficher_Rot_ou_CLV() {
  Effacer_zone(106,4,3); display.setCursor(106,4);
  if (Mode_en_cours == 5) Aff_TEXTE_EEPROM(179,3); else Aff_TEXTE_EEPROM(614,3); display.update();}
  
void Incrementer_Decrementer(byte MAXI) {// La valeur minimale est toujours 1.
  if (Sens_positif) {NB_a_incrementer_decrementer++; 
                     if (NB_a_incrementer_decrementer > MAXI) NB_a_incrementer_decrementer = 1;}
  else { if (NB_a_incrementer_decrementer > 1) NB_a_incrementer_decrementer--;
         else NB_a_incrementer_decrementer = MAXI;}}
         
void Traite_Item_Menus() {
   switch (Mode_en_cours) {
     case 1 : {display.clear(); Affiche_Ecran_MENU(); MENU_EXPLOITER(); break;}
     case 5 : {MENU_MOUVOIR(); break;}
     case 2 : {MENU_POSTURES(); break;}
     case 3 : {MENU_des_OPTIONS(); break;}
     case 4 :  MENU_des_DONNEES();}}
     
void Teste_Touche_LASER() {
     if (BP == 15) {LASER = !LASER; if (LASER) Envoyer_Consigne(29); else Envoyer_Consigne(30);}}
  
void Afficher_le_NB_fois() {
  Effacer_zone(66,55,1); display.setCursor(18,55); Aff_TEXTE_EEPROM(90,8);
  display.print(NB_Repetition_Mvt); Aff_TEXTE_EEPROM(617,5); display.update();}

//--------------------------------- Gestion du clavier à 16 touches --------------------------------
void Tester_le_clavier() {
   BP = 0; // Indique une non utilisation du clavier.
  Teste_Colonne(C1); Teste_Colonne(C2); Teste_Colonne(C3); Teste_Colonne(C4);}

void Teste_Colonne(byte Colonne) {
  digitalWrite(Colonne, HIGH); // Passe la colonne au +5Vcc.
  CNA = analogRead(Mesure_U_clavier);
  if (CNA > 90) {// Codage du BP sollicité.
      digitalWrite(LED_Clavier,HIGH); // Signale l'enfoncement d'une touche.
      BP = Colonne + 6; if (CNA > 286) BP = BP - 4; if (CNA > 511) BP = BP - 4; if (CNA > 718) BP = BP - 4;
      while (analogRead(Mesure_U_clavier) > 90) delay (5); delay (10); // Attendre le relâcher.
      digitalWrite(LED_Clavier,LOW);} // Signale le relâcher de la touche.
  digitalWrite(Colonne, LOW);} // Repasse la colonne à zéro volts.
  
void ATTENDRE_un_BP() {
    while (BP == 0) {
           Tester_le_clavier(); digitalWrite(LED_Clavier,!digitalRead(LED_Clavier)); delay(50);}
    digitalWrite(LED_Clavier,LOW);}

//--------------------------------------------------------------------------------------------------
void BIP() {
  pinMode(BUZZER, OUTPUT); tone(BUZZER,3000); digitalWrite(LED_Erreur,HIGH);
  delay(100); noTone(BUZZER); digitalWrite(LED_Erreur,LOW); pinMode(BUZZER, INPUT);}

void Attendre_une_chaine_sur_le_Maitre() {// Cette procédure attend une chaine sur la ligne USB.
  Chaine_Memorisee = ""; // Vider le tampon pour une nouvelle capture.
  Sentinelle_detectee = false;  // Sentinelle non reçue.
  Compteur_caracteres = 0;
  while (!Sentinelle_detectee) {
    if (Serial.available()) {Caractere = Serial.read(); // Puiser un caractère dans le FIFO.
    Sentinelle_detectee = (Caractere == '*'); // Si sentinelle détectée, informer la boucle de réception.
    // Puis concaténer ce caractère SAUF la sentinelle :
    if ((Compteur_caracteres < LGR_chaine) && (!Sentinelle_detectee)) 
       {Chaine_Memorisee = Chaine_Memorisee + Caractere; Compteur_caracteres++;}}}}    

void Demande_Information(byte Code_consigne) {
  Verifier_OK = false; Transmettre_le_code_consigne(Code_consigne);}

void Transmettre_le_code_consigne(byte Code_consigne) {  
  Serial.print(Code_consigne); Serial.print('*');
  //---------------- Efface le Code en bas à droite sous conditions ----------------
  if ((BP != 4) && (BP != 0) && (Code_consigne != 84))
     {Efface_le_code(); display.setCursor(115,54); display.print(Code_consigne); display.update();}
  //--------------------------------------------------------------------------------
  Attendre_une_chaine_sur_le_Maitre();
  if (Verifier_OK) AFF_ACR();}
      
void AFF_ACR() {
  if ((Chaine_Memorisee != "OK") && (Chaine_Memorisee.charAt(0) != 'P')) BIP(); 
       Effacer_ACR(); display.setCursor(4,4); display.print(Chaine_Memorisee); display.update();}

void Effacer_ACR() {Effacer_zone(4, 4, 3); display.update();}
      
void Efface_le_code() {Effacer_zone(114,53,2);}
      
void Envoyer_Consigne(byte Code_consigne) {
  Verifier_OK = true; Transmettre_le_code_consigne(Code_consigne);}

//--------------------------------- Routines pour les affichages -----------------------------------
void Aff_TEXTE_EEPROM(int PTR, byte Longueur) {
  for (byte I=0; I<Longueur; I++) {display.print(char(EEPROM.read(PTR))); PTR++;}; }
  
void Effacer_zone(byte X, byte Y, byte NB_caracteres) {
  byte I = 10; 
  for (byte J=0; J < I; J++)
       display.drawLine(X, Y + J, X + (NB_caracteres * 6), Y + J, BLACK);}

void CADRE_UP() {display.drawRect(0, 0, 128, 16, WHITE);}
void CADRE() {display.drawRect(0, 16, 128, 48, WHITE);}
void Affiche_cadre_du_haut(byte X) {CADRE_UP(); display.setCursor(X,4);}

